from __future__ import annotations

from importlib import import_module

import awkward as ak

from hml.observables import Observable, parse_observable


class Set:
    """A set of observables.

    Set is a 1D representation of an event. It contains a list of observables.
    It is usually used as input to approaches like CutAndCount and
    ToyMultilayerPerceptron (MLP).

    Parameters
    ----------
    *observables : list[Observable | str]
        A list of observables or their names.
    """

    def __init__(self, observables: list[str | Observable]):
        self.observables = self._init_observables(observables)
        self._values = None

    def _init_observables(self, observables: list[str | Observable]):
        output = []
        for obs in observables:
            if isinstance(obs, str):
                output.append(parse_observable(obs))
            else:
                output.append(obs)

        return output

    def read(self, events):
        values = []
        for obs in self.observables:
            obs.read(events)

            if len(obs.shape) == 2:
                # Maybe a single or collective observable
                if obs.shape[-1] != "1":
                    raise ValueError

                value = obs.value

            if len(obs.shape) == 3:
                # Should be the delta_r observable
                if obs.shape[1] == "1" and obs.shape[2] == "1":
                    value = obs.value[:, :, 0]

            values.append(value)

        self._values = ak.concatenate(values, axis=1)

        return self

    @property
    def names(self):
        return [i.name for i in self.observables]

    @property
    def values(self):
        return self._values

    @property
    def config(self):
        return {
            "observable_configs": {
                i: {
                    "class_name": obs.__class__.__name__,
                    "config": obs.config,
                }
                for i, obs in enumerate(self.observables)
            },
        }

    @classmethod
    def from_config(cls, config):
        observables = []
        module = import_module("hml.observables")

        for i_config in config["observable_configs"].values():
            class_name = i_config["class_name"]
            class_config = i_config["config"]
            class_ = getattr(module, class_name)
            observables.append(class_.from_config(class_config))

        return cls(observables)



from __future__ import annotations

import re

from .collective import Collective
from .physics_object import PhysicsObject
from .single import Single, is_single


def is_nested(object_: PhysicsObject | str) -> bool:
    """Check if an object is a nested physics object"""
    if isinstance(object_, PhysicsObject):
        return isinstance(object_, Nested)

    return bool(re.match(r"^[a-zA-Z]+\d*:?\d*\.[a-zA-Z]+\d*:?\d*$", object_))


class Nested(PhysicsObject):
    """A nested physics object"""

    def __init__(
        self,
        main: PhysicsObject | str,
        sub: PhysicsObject | str,
    ) -> None:
        self._main = self._init_object(main)
        self._sub = self._init_object(sub)

    def _init_object(self, object_: PhysicsObject | str) -> PhysicsObject:
        if isinstance(object_, PhysicsObject):
            return object_

        elif is_single(object_):
            return Single.from_name(object_)

        else:
            return Collective.from_name(object_)

    @classmethod
    def from_name(cls, name: str) -> Nested:
        if "." in name:
            main, sub = name.split(".")
            return cls(main, sub)

        raise ValueError(f"Invalid name '{name}' for a nested physics object")

    @property
    def main(self) -> PhysicsObject:
        return self._main

    @property
    def sub(self) -> PhysicsObject:
        return self._sub

    @property
    def branch(self) -> str:
        return f"{self.main.branch}.{self.sub.branch}"

    @property
    def slices(self) -> list[slice]:
        return [*self.main.slices, *self.sub.slices]

    @property
    def name(self) -> str:
        return f"{self.main.name}.{self.sub.name}"

    @property
    def config(self) -> dict:
        return {"main": self.main.name, "sub": self.sub.name}



from __future__ import annotations

import json
import zipfile
from functools import reduce
from io import BytesIO

import awkward as ak
import numpy as np
from matplotlib import pyplot as plt
from sklearn.model_selection import train_test_split

from hml.approaches import Cut
from hml.representations import Image


class ImageDataset:
    def __init__(self, representation: Image):
        self.image = representation
        self._samples = [] if self.image.been_pixelated else [[], []]
        self._targets = []
        self.been_split = False
        self.train = None
        self.test = None
        self.val = None
        self.seed = None

        self._data = None
        self._been_read = None

    def read(self, events, target, cuts: list[str | Cut] | None = None):
        self.image.read(events)
        if cuts is not None:
            compiled_cuts = []
            for i in cuts:
                if isinstance(i, str):
                    compiled_cuts.append(Cut(i).read(events).value)
                else:
                    compiled_cuts.append(i.read(events).value)
            mask = reduce(np.logical_and, compiled_cuts)

            if self.image.been_pixelated:
                image_values = self.image.values[mask]
            else:
                image_values = [self.image.values[0][mask], self.image.values[1][mask]]
        else:
            image_values = self.image.values

        if self.image.status:
            if isinstance(self._samples, list):
                self._samples = image_values
            else:
                self._samples = ak.concatenate([self._samples, image_values])

            if self.image.been_pixelated:
                if isinstance(self._targets, list):
                    self._targets = ak.values_astype(
                        ak.Array([target] * len(image_values)), "int32"
                    )
                else:
                    self._targets = ak.concatenate(
                        [
                            self._targets,
                            ak.values_astype(
                                ak.Array([target] * len(image_values)), "int32"
                            ),
                        ]
                    )

            else:
                if isinstance(self._targets, list):
                    self._targets = ak.values_astype(
                        ak.Array([target] * len(image_values[0])), "int32"
                    )
                else:
                    self._targets = ak.concatenate(
                        [
                            self._targets,
                            ak.values_astype(
                                ak.Array([target] * len(image_values[0])), "int32"
                            ),
                        ]
                    )
                # self._samples[0].append(self.image.values[0])
                # self._samples[1].append(self.image.values[1])

        # if cut is not None:
        #     self._cut = cut
        #     self._targets = self._targets[cut]
        #     if self.image.been_pixelated:
        #         self._samples = self._samples[cut]
        #     else:
        #         self._samples = list(self._samples)
        #         self._samples[0] = self._samples[0][cut]
        #         self._samples[1] = self._samples[1][cut]

    def split(self, train, test, val=None, seed=None):
        train *= 10
        test *= 10
        samples = self.samples
        targets = self.targets
        self.seed = seed

        if val is None:
            if train + test != 10:
                raise ValueError("train + test must be 1")

        x_train, x_test, y_train, y_test = train_test_split(
            samples,
            targets,
            test_size=test / 10,
            random_state=seed,
        )

        if val is not None:
            val *= 10
            if train + test + val != 10:
                raise ValueError("train + test + val must be 1")

            x_train, x_val, y_train, y_val = train_test_split(
                x_train,
                y_train,
                test_size=val / (train + val),
                random_state=seed,
            )

            self.val = ImageDataset(representation=self.image)
            self.val._samples = x_val
            self.val._targets = y_val

        self.train = ImageDataset(representation=self.image)
        self.train._samples = x_train
        self.train._targets = y_train
        self.test = ImageDataset(representation=self.image)
        self.test._samples = x_test
        self.test._targets = y_test

        self.been_split = True

    def save(self, filepath="dataset.ds"):
        configs = self.config
        configs_json = json.dumps(configs)

        npz_data = BytesIO()
        np.savez(npz_data, samples=self.samples, targets=self.targets)
        npz_data.seek(0)

        with zipfile.ZipFile(filepath, "w") as zf:
            zf.writestr("configs.json", configs_json)
            zf.writestr("data.npz", npz_data.read())

            if self.been_split:
                npz_train = BytesIO()
                self.train.save(npz_train)
                npz_train.seek(0)
                zf.writestr("train.ds", npz_train.read())

                npz_test = BytesIO()
                self.test.save(npz_test)
                npz_test.seek(0)
                zf.writestr("test.ds", npz_test.read())

                if self.val is not None:
                    npz_val = BytesIO()
                    self.val.save(npz_val)
                    npz_val.seek(0)
                    zf.writestr("val.ds", npz_val.read())

    @classmethod
    def load(cls, filepath, lazy=True):
        zf = zipfile.ZipFile(filepath)
        # Extract and read configs JSON
        with zf.open("configs.json") as json_file:
            configs = json.load(json_file)

        dataset = cls.from_config(configs)
        dataset._filepath = filepath
        dataset._data = zf.open("data.npz")

        # Extract and load .npz file
        if not lazy:
            with dataset._data as split_zf:
                split_data = np.load(BytesIO(split_zf.read()))
                dataset._samples = split_data["samples"]
                dataset._targets = split_data["targets"]
                dataset._been_read = True
        else:
            dataset._been_read = False

        # Extract and load train, test, and val .npz files
        if configs["been_split"]:
            dataset.train = cls.load(zf.open("train.ds"), lazy=lazy)
            dataset.test = cls.load(zf.open("test.ds"), lazy=lazy)

            if "val.ds" in [i.filename for i in zf.filelist]:
                dataset.val = cls.load(zf.open("val.ds"), lazy=lazy)

        return dataset

    @property
    def samples(self):
        if self._been_read is False and self._data is not None:
            split_data = np.load(BytesIO(self._data.read()))
            self._samples = split_data["samples"]
            self._data.seek(0)

            if np.array(self._samples).size > 0 and np.array(self._targets).size > 0:
                self._been_read = True

        if self.image.been_pixelated:
            return np.array(self._samples, dtype=np.float32)
        else:
            height = ak.from_iter(self._samples[0])
            width = ak.from_iter(self._samples[1])

            # 1D non-pixelated data should come from a loaded dataset
            if height.ndim == 1:
                height = ak.to_numpy(height)
                width = ak.to_numpy(width)

            # 2D non-pixelated data should come from an event loop
            else:
                height = ak.to_numpy(ak.flatten(height))
                width = ak.to_numpy(ak.flatten(width))

            return height, width

    @property
    def targets(self):
        if self._been_read is False and self._data is not None:
            split_data = np.load(BytesIO(self._data.read()))
            self._targets = split_data["targets"]
            self._data.seek(0)

            if np.array(self._samples).size > 0 and np.array(self._targets).size > 0:
                self._been_read = True

        return np.array(self._targets, dtype=np.int32)

    @property
    def features(self):
        return {
            "height": self.image.height.name,
            "width": self.image.width.name,
            "channel": self.image.channel.name if self.image.channel else None,
        }

    def show(
        self,
        limits=None,
        show_pixels=False,
        grid=True,
        norm=None,
        n_samples=-1,
        target=None,
    ):
        if target is not None and self.image.been_pixelated:
            samples = self.samples[np.squeeze(self.targets) == target]

        else:
            samples = self.samples

        plt.figure()

        if not self.image.been_pixelated:
            plt.scatter(
                x=samples[1][:n_samples],
                y=samples[0][:n_samples],
                c="k",
                s=10,
                marker="o",
            )

            # Restrict the range of axes.
            if limits is not None:
                plt.xlim(limits[0])
                plt.ylim(limits[1])

        else:
            # Use pcolormesh instead of imshow so that the actual values of bins
            # are shown.
            plt.pcolormesh(
                self.image.w_bins,
                self.image.h_bins,
                samples[:n_samples].sum(0).T,
                norm=norm,
            )

            # Turn on colorbar to show the range of values.
            plt.colorbar()

            if show_pixels:
                # Turn off ticks and labels.
                plt.tick_params(
                    bottom=False,
                    left=False,
                    labelbottom=False,
                    labelleft=False,
                )
                plt.xticks(self.image.w_bins)
                plt.yticks(self.image.h_bins)

            # Set aspect ratio to be equal so that every pixel is square.
            # Ignore the actual values of bins since here is just a schematic.
            plt.gca().set_aspect("equal")

        # Turn on grid to show pixel boundaries.
        if grid:
            plt.grid(alpha=0.5)

        plt.show()

    @property
    def config(self):
        config = self.image.config
        config.update(
            {
                "class_name": self.__class__.__name__,
                "been_split": self.been_split,
                "seed": self.seed,
            }
        )

        return config

    @classmethod
    def from_config(cls, config):
        image = Image.from_config(config)

        instance = cls(image)
        instance.been_split = config["been_split"]
        instance.seed = config["seed"]

        return instance



from __future__ import annotations

import re

from .physics_object import PhysicsObject


def is_single(object_: PhysicsObject | str) -> bool:
    """Check if an object is a single physics object"""
    if isinstance(object_, PhysicsObject):
        return isinstance(object_, Single)

    return bool(re.match(r"^[a-zA-Z]+\d+$", object_))


class Single(PhysicsObject):
    """A single physics object"""

    def __init__(self, branch: str, index: int) -> None:
        self._branch = branch
        self._index = index

    @classmethod
    def from_name(cls, name: str) -> Single:
        if match_ := re.match(r"^([a-zA-Z]+)(\d+)$", name.strip()):
            branch, index = match_.groups()
            return cls(branch, int(index))

        raise ValueError(f"Invalid name '{name}' for a single physics object")

    @property
    def index(self) -> int:
        return self._index

    @property
    def branch(self) -> str:
        return self._branch

    @property
    def slices(self) -> list[slice]:
        return [slice(self.index, self.index + 1)]

    @property
    def name(self) -> str:
        return f"{self.branch}{self.index}"

    @property
    def config(self) -> dict:
        return {"branch": self.branch, "index": self.index}



from __future__ import annotations

from keras import ops
from keras.metrics import FalsePositives, Metric, TruePositives


class MaxSignificance(Metric):
    def __init__(
        self,
        thresholds=None,
        class_id=1,
        name="max_significance",
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self.class_id = class_id
        self.tp = TruePositives(thresholds=thresholds, dtype=dtype)
        self.fp = FalsePositives(thresholds=thresholds, dtype=dtype)

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_t_ndim = ops.ndim(ops.squeeze(y_true))
        y_p_ndim = ops.ndim(ops.squeeze(y_pred))

        y_true = ops.argmax(y_true, -1) if y_t_ndim == 2 else y_true
        y_pred = ops.take(y_pred, self.class_id, -1) if y_p_ndim == 2 else y_pred

        self.tp.update_state(y_true, y_pred, sample_weight)
        self.fp.update_state(y_true, y_pred, sample_weight)

    def result(self):
        s = self.tp.result()
        b = self.fp.result()
        significance = ops.divide(s, ops.sqrt(s + b))
        max_significance = ops.max(significance)
        return max_significance


def max_significance(): ...



from __future__ import annotations

from keras import ops
from keras.metrics import Metric, SpecificityAtSensitivity


class RejectionAtEfficiency(Metric):
    def __init__(
        self,
        efficiency,
        num_thresholds=200,
        class_id=None,
        name="rejection_at_efficiency",
        dtype=None,
    ):
        super().__init__(name=name, dtype=dtype)
        self.specificity_at_sensitivity = SpecificityAtSensitivity(
            sensitivity=efficiency,
            num_thresholds=num_thresholds,
            class_id=class_id,
            dtype=dtype,
        )

    def update_state(self, y_true, y_pred, sample_weight=None):
        if ops.shape(y_true) != ops.shape(y_pred):  # -> not allowed in non-eager mode
            n_classes = ops.shape(y_pred)[-1]
            y_true = ops.one_hot(ops.squeeze(y_true), n_classes, dtype="int32")

        self.specificity_at_sensitivity.update_state(y_true, y_pred, sample_weight)

    def result(self):
        fpr = ops.subtract(
            1, self.specificity_at_sensitivity.result()
        )  # -> not allowed in non-eager mode
        rejection = ops.divide(1, fpr)
        return rejection


def rejection_at_efficiency(): ...



from __future__ import annotations

import keras
from keras import ops

from hml.operations import ops_histogram_fixed_width, ops_unique

from .cut_layer import CutLayer


@keras.saving.register_keras_serializable()
class CutAndCount(keras.Model):
    def __init__(
        self,
        n_observables: int,
        n_bins: int = 50,
        topology="parallel",
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.n_observables = n_observables
        self.topology = topology
        self.n_bins = n_bins
        self.cut_layers = [CutLayer(feature_id=i) for i in range(n_observables)]

    def train_step(self, data):
        x, y = data
        y_pred = self(x, y)  # (N, 2)
        loss = self.compute_loss(y=y, y_pred=y_pred)

        for metric in self.metrics:
            if metric.name == "loss":
                metric.update_state(loss)
            else:
                metric.update_state(y, y_pred)

        return {m.name: m.result() for m in self.metrics}

    def call(self, x, y=None):
        if self.topology == "parallel":
            y_pred = ops.squeeze(self.parallel_call(x, y))  # (N,)
        elif self.topology == "sequential":
            y_pred = ops.squeeze(self.sequential_call(x, y))  # (N,)
        else:
            raise NotImplementedError()

        return ops.one_hot(ops.cast(y_pred, "int32"), num_classes=2)

    def parallel_call(self, x, y=None):
        y_preds = []

        for cut_layer in self.cut_layers:
            ix = ops.take(x, cut_layer.feature_id, axis=-1)

            if y is not None:
                cut_left, cut_right, case = self.find_best_cut(ix, y)
                cut_layer._cut_left.assign(cut_left)
                cut_layer._cut_right.assign(cut_right)
                cut_layer._case.assign(case)

            y_pred = cut_layer(ix)
            y_preds.append(y_pred)

        return keras.layers.Multiply()(y_preds)

    def sequential_call(self, x, y=None, show_cut_mark=False):
        y_pred = ops.ones(ops.shape(x)[0])
        mask = y_pred
        for cut_layer in self.cut_layers:
            ix = ops.take(x, cut_layer.feature_id, axis=-1)

            if y is not None:
                masked_x = ops.take(ix, ops.squeeze(ops.where(mask >= 0), 0))
                masked_y = ops.take(
                    y, ops.squeeze(ops.where(mask >= 0), 0), 0
                )  # should keep y as original
                cut_left, cut_right, case = self.find_best_cut(masked_x, masked_y)
                cut_layer._cut_left.assign(cut_left)
                cut_layer._cut_right.assign(cut_right)
                cut_layer._case.assign(case)

            mask = cut_layer.apply_cut(ix, -1)
            y_pred = ops.logical_and(y_pred, ops.where(mask < 0, 0.0, 1.0))
            y_pred = ops.cast(y_pred, "float32")

        return y_pred

    def find_best_cut(self, x, y):
        if ops.ndim(y) == 2:
            if ops.ndim(ops.squeeze(y)) == 2:
                iy = ops.argmax(y, -1)
            else:
                iy = ops.squeeze(y)
        else:
            iy = y

        x_min = ops.min(x)
        x_max = ops.max(x)
        bin_edges = ops.linspace(x_min, x_max, self.n_bins + 1)

        is_bkg = ops.squeeze(ops.where(ops.equal(iy, 0)), 0)
        is_sig = ops.squeeze(ops.where(ops.equal(iy, 1)), 0)

        x0 = ops.take(x, is_bkg)
        x1 = ops.take(x, is_sig)

        hist0 = ops_histogram_fixed_width(x0, [x_min, x_max], self.n_bins)
        hist1 = ops_histogram_fixed_width(x1, [x_min, x_max], self.n_bins)

        curr_case0 = ops.greater(hist0, hist1)[:-1]  # type: ignore
        next_case0 = ops.greater(hist0, hist1)[1:]  # type: ignore
        is_on_boundary0 = ops.logical_xor(curr_case0, next_case0)
        is_not_empty0 = hist0[:-1] > 0  # type: ignore
        is_candidate0 = ops.logical_and(is_on_boundary0, is_not_empty0)
        candidate_indices0 = ops.add(ops.squeeze(ops.where(is_candidate0), 0), 1)
        candidates0 = ops.take(bin_edges, candidate_indices0)

        curr_case1 = ops.greater(hist1, hist0)[:-1]  # type: ignore
        next_case1 = ops.greater(hist1, hist0)[1:]  # type: ignore
        is_on_boundary1 = ops.logical_xor(curr_case1, next_case1)
        is_not_empty1 = hist1[:-1] > 0  # type: ignore
        is_candidate1 = ops.logical_and(is_on_boundary1, is_not_empty1)
        candidate_indices1 = ops.add(ops.squeeze(ops.where(is_candidate1), 0), 1)
        candidates1 = ops.take(bin_edges, candidate_indices1)

        candidates = ops_unique(ops.concatenate([candidates0, candidates1], 0))
        candidates = ops.cond(
            ops.equal(ops.shape(candidates)[0], 1),
            lambda: ops.append(candidates, x_max),
            lambda: candidates,
        )
        i, j = ops.meshgrid(candidates, candidates)  # type: ignore
        i, j = i[i < j], j[i < j]
        candidate_pairs = ops.stack([i, j], 1)
        losses_and_cases = ops.vectorized_map(
            lambda pair: self._get_min_loss_and_case(x, y, pair, self.compute_loss),
            candidate_pairs,
        )
        min_index = ops.argmin(ops.take(losses_and_cases, 0, 1))
        # min_loss = losses_and_cases[min_index, 0]  # type: ignore
        min_case = losses_and_cases[min_index, 1]  # type: ignore

        lower = candidate_pairs[min_index, 0]  # type: ignore
        upper = candidate_pairs[min_index, 1]  # type: ignore

        return lower, upper, min_case

    def _get_min_loss_and_case(self, x, y, candidate_pair, loss_fn):  # pragma: no cover
        # I: (n_samples,), (n_samples, ?), (2,)
        cut0 = candidate_pair[0]
        cut1 = candidate_pair[1]

        # For each pair, calculate the loss of the four cases
        # Case 0: signal on the left
        on_left = ops.less_equal(x, cut0)  # (n_samples,)
        y_pred0 = ops.where(on_left, 1.0, 0.0)  # (n_samples,)
        y_pred0 = ops.one_hot(ops.cast(y_pred0, "int32"), num_classes=2)
        loss0 = loss_fn(y=y, y_pred=y_pred0)

        # Case 1: signal on the right
        on_right = ops.greater_equal(x, cut0)
        y_pred1 = ops.where(on_right, 1.0, 0.0)
        y_pred1 = ops.one_hot(ops.cast(y_pred1, "int32"), num_classes=2)
        loss1 = loss_fn(y=y, y_pred=y_pred1)

        # Case 2: signal in the middle
        in_middle = ops.logical_and(x >= cut0, x <= cut1)
        y_pred2 = ops.where(in_middle, 1.0, 0.0)
        y_pred2 = ops.one_hot(ops.cast(y_pred2, "int32"), num_classes=2)
        loss2 = loss_fn(y=y, y_pred=y_pred2)

        # Case 3: signal on both sides
        on_both_sides = ops.logical_or(x <= cut0, x >= cut1)
        y_pred3 = ops.where(on_both_sides, 1.0, 0.0)
        y_pred3 = ops.one_hot(ops.cast(y_pred3, "int32"), num_classes=2)
        loss3 = loss_fn(y=y, y_pred=y_pred3)

        # The cases 0, 1, 2, 3 are corresponding to the four possible cases:
        # left, right, middle, both sides
        min_case = ops.argmin([loss0, loss1, loss2, loss3])
        min_loss = ops.min([loss0, loss1, loss2, loss3])

        return ops.append(min_loss, min_case)  # (2,)

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "n_observables": self.n_observables,
                "n_bins": self.n_bins,
                "topology": self.topology,
            }
        )

        return config



from __future__ import annotations

import awkward as ak

from ..operations import branch_to_momentum4d, constituents_to_momentum4d
from ..physics_objects import PhysicsObject, is_collective, is_multiple, is_single
from ..physics_objects import parse_physics_object as parse_object


class Observable:
    def __init__(
        self,
        physics_object: str | PhysicsObject | None = None,
        class_name: str | None = None,
        supported_objects: list[str | PhysicsObject] | None = None,
    ) -> None:
        self._physics_object = self._init_object(physics_object)
        self._class_name = self._init_class_name(class_name)
        self._supported_objects = self._init_supported_objects(supported_objects)
        self._validate_physics_object()

    def _init_object(self, object_: str | PhysicsObject | None) -> PhysicsObject | None:
        if isinstance(object_, PhysicsObject):
            return object_

        elif isinstance(object_, str):
            return parse_object(object_)

        else:
            return

    def _init_class_name(self, class_name: str | None) -> str:
        return class_name if class_name else self.__class__.__name__

    def _init_supported_objects(
        self, supported_objects: list[str | PhysicsObject] | None
    ) -> list[PhysicsObject] | None:
        if supported_objects is None:
            return

        output = []
        for obj in supported_objects:
            if isinstance(obj, PhysicsObject):
                output.append(obj.__class__.__name__.lower())
            else:
                output.append(obj.lower())

        return output

    def _validate_physics_object(self) -> None:
        if self.supported_objects is None:
            return

        elif self.physics_object is None and self.supported_objects is not None:
            raise ValueError

        elif "multiple" not in self.supported_objects:
            if (
                self.physics_object.__class__.__name__.lower()
                not in self.supported_objects
            ):
                raise ValueError

        else:
            if not is_multiple(self.physics_object, self.supported_objects):
                raise ValueError

    def __eq__(self, other: str | Observable) -> bool:
        if isinstance(other, str):
            other = self.from_name(other)

        return (
            self.physics_object == other.physics_object
            and self.__class__ == other.__class__
        )

    def __repr__(self) -> str:
        return f"{self.name}: {self.value.type!s}"

    @property
    def physics_object(self) -> PhysicsObject | None:
        return self._physics_object

    @property
    def class_name(self) -> str:
        return self._class_name

    @property
    def supported_objects(self) -> list[str] | None:
        return self._supported_objects

    @property
    def value(self) -> ak.Array:
        if not hasattr(self, "_value"):
            self._value = ak.Array([])

        try:
            return ak.to_regular(self._value, axis=None)
        except Exception:
            return self._value

    @property
    def name(self) -> str:
        if self.physics_object:
            return f"{self.physics_object.name}.{self.class_name}"
        else:
            return self.class_name

    @property
    def config(self) -> dict:
        return {
            "physics_object": self.physics_object.name,
            "class_name": self.class_name,
        }

    def read(self, events) -> Observable:
        if "multiple" in self.supported_objects:
            raise NotImplementedError

        all_keys = {i.lower(): i for i in events.keys(full_paths=False)}
        branch = self.physics_object.branch.lower()
        slices = self.physics_object.slices

        if branch is None:
            raise ValueError(f"Branch {self.physics_object.branch} not found")

        if is_single(self.physics_object) or is_collective(self.physics_object):
            if f"{branch}.{self.class_name.lower()}" in all_keys:
                key = all_keys[f"{branch}.{self.class_name.lower()}"]
                value = events[key].array()

            else:
                array = branch_to_momentum4d(events, all_keys[branch])
                value = getattr(array, self.class_name.lower())

        else:
            array = constituents_to_momentum4d(events, all_keys[branch])
            value = getattr(array, self.class_name.lower())

        if len(slices) == 1:
            value = value[:, slices[0]]
        else:
            value = value[:, slices[0], slices[1]]

        for i, slice_ in enumerate(slices):
            if slice_.stop is not None:
                start = slice_.start if slice_.start is not None else 0
                required_length = slice_.stop - start

                if i + 1 == len(slices) and ak.any(
                    ak.num(value, i + 1) < required_length
                ):
                    value = ak.pad_none(value, required_length, axis=i + 1)

                else:
                    n_missing = required_length - ak.num(value, i + 1)
                    if ak.sum(n_missing) > 0:
                        pad = ak.unflatten(
                            ak.Array([[]] * ak.sum(n_missing)), n_missing
                        )
                        value = ak.concatenate([value, pad], axis=i + 1)

        self._value = value

        return self

    @classmethod
    def from_name(cls, name: str, **kwargs) -> "Observable":
        *parts, class_name = name.split(".")
        physics_object = ".".join(parts) if len(parts) > 0 else None

        if "class_name" in kwargs:
            class_name = kwargs["class_name"]
            del kwargs["class_name"]

        return cls(physics_object=physics_object, class_name=class_name, **kwargs)

    @classmethod
    def from_config(cls, config: dict) -> "Observable":
        return cls(**config)

    @property
    def shape(self) -> str:
        *parts, _ = str(self.value.type).split(" * ")
        return tuple(parts)

    @property
    def dtype(self) -> type:
        *_, dtype = str(self.value.type).split(" * ")
        return dtype



from __future__ import annotations

from functools import reduce

import awkward as ak
import vector

from hml.physics_objects.physics_object import PhysicsObject

from ..operations import branch_to_momentum4d
from .observable import Observable

vector.register_awkward()


class InvariantMass(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "multiple"]
        super().__init__(physics_object, class_name, supported_objects)

    def read(self, events):
        all_keys = {i.lower(): i for i in events.keys(full_paths=False)}

        momenta = []
        for obj in self.physics_object.all:
            momentum4d = branch_to_momentum4d(events, all_keys[obj.branch.lower()])
            padded_momentum4d = ak.pad_none(momentum4d[:, obj.slices[0]], 1)
            momenta.append(padded_momentum4d)

        total = reduce(lambda x, y: x + y, momenta)
        self._value = total.mass

        return self



from __future__ import annotations

from .collective import Collective, is_collective
from .nested import Nested
from .physics_object import PhysicsObject
from .single import Single, is_single


def is_multiple(
    object_: PhysicsObject | str,
    supported_types: list[PhysicsObject | str] | None = None,
) -> bool:
    """Check if an object is a multiple physics object"""
    if isinstance(object_, str):
        try:
            object_ = Multiple.from_name(object_)
        except Exception:
            return False

    if supported_types is None:
        return True

    supported_types = [
        i.lower() if isinstance(i, str) else i.__class__.__name__.lower()
        for i in supported_types
    ]

    for obj in object_.all:
        if obj.__class__.__name__.lower() not in supported_types:
            return False

    return True


class Multiple(PhysicsObject):
    """A multiple physics object"""

    def __init__(self, all: list[PhysicsObject | str]) -> None:
        self._all = self._init_all(all)

    def _init_all(self, objects: list[PhysicsObject | str]) -> list[PhysicsObject]:
        output = []
        for obj in objects:
            if isinstance(obj, PhysicsObject):
                output.append(obj)

            elif is_single(obj):
                output.append(Single.from_name(obj))

            elif is_collective(obj):
                output.append(Collective.from_name(obj))

            else:
                output.append(Nested.from_name(obj))

        return output

    @classmethod
    def from_name(cls, name: str) -> Multiple:
        if "," in name:
            all_ = name.split(",")
            return cls(all_)

        raise ValueError(f"Invalid name '{name}' for a multiple physics object")

    @property
    def all(self) -> list[PhysicsObject]:
        return self._all

    @property
    def branch(self) -> list[str]:
        return [obj.branch for obj in self.all]

    @property
    def slices(self) -> list[list[slice]]:
        return [i.slices for i in self.all]

    @property
    def name(self) -> str:
        return ",".join(obj.name for obj in self.all)

    @property
    def config(self) -> dict:
        return {"all": [i.name for i in self.all]}



from __future__ import annotations

import awkward as ak
import numpy as np

from hml.physics_objects.physics_object import PhysicsObject

from .kinematics import Eta, Phi
from .observable import Observable


class AngularDistance(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested", "multiple"]
        super().__init__(physics_object, class_name, supported_objects)
        assert (
            len(self.physics_object.all) == 2
        ), "Two physics objects are required for angular distance"

    def read(self, events) -> None:
        obj0_eta = Eta(self.physics_object.all[0]).read(events).value
        obj0_phi = Phi(self.physics_object.all[0]).read(events).value

        obj1_eta = Eta(self.physics_object.all[1]).read(events).value
        obj1_phi = Phi(self.physics_object.all[1]).read(events).value

        if obj0_eta.ndim == 3:
            obj0_eta = ak.flatten(obj0_eta, axis=-1)
            obj0_phi = ak.flatten(obj0_phi, axis=-1)

        if obj1_eta.ndim == 3:
            obj1_eta = ak.flatten(obj1_eta, axis=-1)
            obj1_phi = ak.flatten(obj1_phi, axis=-1)

        delta_eta = obj0_eta[:, :, None] - obj1_eta[:, None, :]
        delta_phi = obj0_phi[:, :, None] - obj1_phi[:, None, :]
        self._value = np.sqrt(delta_eta**2 + delta_phi**2)

        return self

    @classmethod
    def from_name(cls, name: str, **kwargs) -> AngularDistance:
        *parts, class_name = name.split(".")
        physics_object = ".".join(parts) if len(parts) > 0 else None

        return cls(physics_object, class_name)



from __future__ import annotations

import keras
from keras import initializers, ops


@keras.saving.register_keras_serializable()
class CutLayer(keras.Layer):
    def __init__(
        self,
        cut: str | None = None,
        count=0,
        cut_left: float = 0.0,
        cut_right: float = 0.0,
        case: int = 0,
        feature_id: int = 0,
        **kwargs,
    ):
        if cut is not None:
            super().__init__(name=cut, **kwargs)
        else:
            super().__init__(**kwargs)
            self.name = self.name.replace("cut_layer", "observable")

        self._cut = cut
        self._count = count
        self._cut_left = self.add_weight(
            shape=(),
            initializer=initializers.Constant(cut_left),
            trainable=False,
        )

        self._cut_right = self.add_weight(
            shape=(),
            initializer=initializers.Constant(cut_right),
            trainable=False,
        )

        self._case = self.add_weight(
            shape=(),
            initializer=initializers.Constant(case),
            trainable=False,
        )

        self.feature_id = feature_id

    @property
    def cut(self):
        if self._cut is not None:
            return self._cut
        else:
            if self.case == 0:
                return f"{self.name} <= {self.cut_left:.5f}"
            elif self.case == 1:
                return f"{self.name} >= {self.cut_left:.5f}"
            elif self.case == 2:
                return f"{self.cut_left:.5f} <= {self.name} <= {self.cut_right:.5f}"
            else:
                return f"{self.name} <= {self.cut_left:.5f} or {self.name} >= {self.cut_right:.5f}"

    @property
    def count(self):
        return int(ops.convert_to_numpy(self._count))

    @property
    def cut_left(self):
        return float(ops.convert_to_numpy(self._cut_left))

    @property
    def cut_right(self):
        return float(ops.convert_to_numpy(self._cut_right))

    @property
    def case(self):
        return int(ops.convert_to_numpy(self._case))

    def call(self, x):
        y_pred = ops.cond(
            ops.equal(self._case, 0),
            lambda: ops.where(x <= self._cut_left, 1.0, 0.0),
            lambda: ops.cond(
                ops.equal(self._case, 1),
                lambda: ops.where(x >= self._cut_left, 1.0, 0.0),
                lambda: ops.cond(
                    ops.equal(self._case, 2),
                    lambda: ops.where(
                        ops.logical_and(x >= self._cut_left, x <= self._cut_right),
                        1.0,
                        0.0,
                    ),
                    lambda: ops.where(
                        ops.logical_or(x <= self._cut_left, x >= self._cut_right),
                        1.0,
                        0.0,
                    ),
                ),
            ),
        )

        return y_pred

    def apply_cut(self, x, cut_mark=-1.0):
        x = ops.cast(x, "float32")
        if ops.ndim(x) == 2:
            x = ops.take(x, self.feature_id, axis=-1)

        y_pred = ops.cond(
            ops.equal(self._case, 0),
            lambda: ops.where(x <= self._cut_left, 1.0, cut_mark),
            lambda: ops.cond(
                ops.equal(self._case, 1),
                lambda: ops.where(x >= self._cut_left, 1.0, cut_mark),
                lambda: ops.cond(
                    ops.equal(self._case, 2),
                    lambda: ops.where(
                        ops.logical_and(x >= self._cut_left, x <= self._cut_right),
                        1.0,
                        cut_mark,
                    ),
                    lambda: ops.where(
                        ops.logical_or(x <= self._cut_left, x >= self._cut_right),
                        1.0,
                        cut_mark,
                    ),
                ),
            ),
        )
        return y_pred

    def compute_output_shape(self, input_shape):
        return input_shape

    def get_config(self):
        config = super().get_config()
        config.update(
            {
                "cut": self._cut,
                "count": self._count,
                "feature_id": self.feature_id,
            }
        )
        return config



from __future__ import annotations

from ..physics_objects import PhysicsObject
from .observable import Observable


class Size(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["collective"]
        super().__init__(physics_object, class_name, supported_objects)

    def read(self, events) -> Observable:
        all_keys = {i.lower(): i for i in events.keys(full_paths=False)}
        branch = self.physics_object.branch.lower()

        if f"{branch}_size" in all_keys:
            key = all_keys[f"{branch}_size"]
            value = events[key].array()

        self._value = value

        return self



from __future__ import annotations

import keras
from keras import Model
from keras.layers import Dense, Input


@keras.saving.register_keras_serializable()
class SimpleMLP(Model):
    def __init__(self, input_shape, name="simple_mlp", **kwargs):
        super().__init__(name=name, **kwargs)
        self.input_shape = input_shape
        self.dense1 = Dense(units=32, activation="relu")
        self.dense2 = Dense(units=64, activation="relu")
        self.dense3 = Dense(units=32, activation="relu")
        self.dense4 = Dense(units=2, activation="softmax")

        self.call(Input(shape=input_shape))

    def call(self, x):
        x = self.dense1(x)
        x = self.dense2(x)
        x = self.dense3(x)
        return self.dense4(x)

    def get_config(self):
        base_config = super().get_config()
        config = {"input_shape": self.input_shape}
        return {**base_config, **config}



from __future__ import annotations

import json
import zipfile
from functools import reduce
from io import BytesIO

import awkward as ak
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

from hml.approaches import Cut
from hml.observables import Observable
from hml.representations import Set


class SetDataset:
    def __init__(self, observables: list[str | Observable]):
        self.set = Set(observables)
        self.been_split = False
        self.seed = None

        self._samples = []
        self._targets = []
        self.train = None
        self.test = None
        self.val = None

        self._data = None
        self._been_read = False

    def read(self, events, target, cuts: list[str | Cut] | None = None):
        self.set.read(events)
        if cuts is not None:
            compiled_cuts = []
            for i in cuts:
                if isinstance(i, str):
                    compiled_cuts.append(Cut(i).read(events).value)
                else:
                    compiled_cuts.append(i.read(events).value)

            mask = reduce(np.logical_and, compiled_cuts)
            set_values = self.set.values[mask]
        else:
            set_values = self.set.values

        if isinstance(self._samples, list):
            self._samples = set_values
        else:
            self._samples = ak.concatenate([self._samples, set_values])

        if isinstance(self._targets, list):
            self._targets = ak.values_astype(
                ak.Array([target] * len(set_values)), "int32"
            )
        else:
            self._targets = ak.concatenate(
                [
                    self._targets,
                    ak.values_astype(ak.Array([target] * len(set_values)), "int32"),
                ]
            )

    def split(self, train, test, val=None, seed=None):
        train *= 10
        test *= 10
        samples = self.samples
        targets = self.targets
        self.seed = seed

        if val is None:
            if train + test != 10:
                raise ValueError("train + test must be 1")

        x_train, x_test, y_train, y_test = train_test_split(
            samples,
            targets,
            test_size=test / 10,
            random_state=seed,
        )

        if val is not None:
            val *= 10
            if train + test + val != 10:
                raise ValueError("train + test + val must be 1")

            x_train, x_val, y_train, y_val = train_test_split(
                x_train,
                y_train,
                test_size=val / (train + val),
                random_state=seed,
            )

            self.val = SetDataset(self.set.observables)
            self.val._samples = x_val
            self.val._targets = y_val

        self.train = SetDataset(self.set.observables)
        self.train._samples = x_train
        self.train._targets = y_train
        self.test = SetDataset(self.set.observables)
        self.test._samples = x_test
        self.test._targets = y_test

        self.been_split = True

    def save(self, filepath="dataset.ds"):
        configs = self.config
        configs_json = json.dumps(configs)

        npz_data = BytesIO()
        np.savez(npz_data, samples=self.samples, targets=self.targets)
        npz_data.seek(0)

        with zipfile.ZipFile(filepath, "w") as zf:
            zf.writestr("configs.json", configs_json)
            zf.writestr("data.npz", npz_data.read())

            if self.been_split:
                npz_train = BytesIO()
                self.train.save(npz_train)
                npz_train.seek(0)
                zf.writestr("train.ds", npz_train.read())

                npz_test = BytesIO()
                self.test.save(npz_test)
                npz_test.seek(0)
                zf.writestr("test.ds", npz_test.read())

                if self.val is not None:
                    npz_val = BytesIO()
                    self.val.save(npz_val)
                    npz_val.seek(0)
                    zf.writestr("val.ds", npz_val.read())

    @classmethod
    def load(cls, filepath, lazy=True):
        zf = zipfile.ZipFile(filepath)
        # Extract and read configs JSON
        with zf.open("configs.json") as json_file:
            configs = json.load(json_file)

        dataset = cls.from_config(configs)
        dataset._filepath = filepath
        dataset._data = zf.open("data.npz")

        # Extract and load .npz file
        if not lazy:
            with dataset._data as split_zf:
                split_data = np.load(BytesIO(split_zf.read()))
                dataset._samples = split_data["samples"]
                dataset._targets = split_data["targets"]
                dataset._been_read = True

        # Extract and load train, test, and val .npz files
        if configs["been_split"]:
            dataset.train = cls.load(zf.open("train.ds"), lazy=lazy)
            dataset.test = cls.load(zf.open("test.ds"), lazy=lazy)

            if "val.ds" in [i.filename for i in zf.filelist]:
                dataset.val = cls.load(zf.open("val.ds"), lazy=lazy)

        return dataset

    @property
    def samples(self):
        if len(self._samples) == 0 and not self._been_read and self._data is not None:
            split_data = np.load(BytesIO(self._data.read()))
            self._samples = split_data["samples"]
            self._data.seek(0)

        if len(self._samples) > 0 and len(self._targets) > 0:
            self._been_read = True

        # return np.array(self._samples, dtype=np.float32)
        nan_float32 = np.array(np.nan, dtype=np.float32)
        return ak.to_numpy(ak.fill_none(self._samples, nan_float32))
        # return ak.to_numpy(self._samples, allow_missing=False)

    @property
    def targets(self):
        if len(self._targets) == 0 and not self._been_read and self._data is not None:
            split_data = np.load(BytesIO(self._data.read()))
            self._targets = split_data["targets"]
            self._data.seek(0)

        if len(self._samples) > 0 and len(self._targets) > 0:
            self._been_read = True

        # return np.array(self._targets, dtype=np.int32)
        nan_float32 = np.array(np.nan, dtype=np.float32)
        return ak.to_numpy(ak.fill_none(self._targets, nan_float32))
        # return ak.to_numpy(self._targets, allow_missing=False)

    @property
    def feature_names(self):
        return self.set.names

    def to_numpy(self):
        return np.hstack([self.samples, self.targets[:, None]])

    def to_pandas(self):
        df = pd.DataFrame(self.samples.tolist(), columns=self.feature_names)
        df["Target"] = self.targets
        return df

    @property
    def config(self):
        config = self.set.config
        config.update(
            {
                "class_name": self.__class__.__name__,
                "been_split": self.been_split,
                "seed": self.seed,
            }
        )

        return config

    @classmethod
    def from_config(cls, config):
        set = Set.from_config(config)

        instance = cls(set.observables)
        instance.been_split = config["been_split"]
        instance.seed = config["seed"]

        return instance

    def show(self, n_feature_per_line=3, n_samples=-1, target=None):
        if n_samples != -1:
            samples = self.samples[:n_samples]
        else:
            samples = self.samples

        n_features = len(self.feature_names)
        plt.figure(figsize=(3 * n_feature_per_line, 3))
        for i, name in enumerate(self.feature_names):
            ax = plt.subplot(
                (n_features + n_feature_per_line - 1) // n_feature_per_line,
                n_feature_per_line,
                i + 1,
            )

            if target is None:
                for i_target in np.unique(self.targets):
                    ax.hist(
                        samples[np.squeeze(self.targets == i_target)][:, i],
                        bins=50,
                        histtype="step",
                        label=f"Target {i_target}",
                    )
            else:
                ax.hist(
                    samples[np.squeeze(self.targets == i_target)][:, i],
                    bins=50,
                    histtype="step",
                    label=f"Target {i_target}",
                )
            ax.set_title(name)

        plt.legend()
        plt.tight_layout()
        plt.show()



import pickle
from pathlib import Path

from keras.saving import load_model

from .cuts import Cut, CutAndCount, CutLayer
from .networks import SimpleCNN, SimpleGNN, SimpleMLP
from .trees import GradientBoostedDecisionTree


def load_approach(filepath):
    filepath = Path(filepath)

    if filepath.suffix == ".keras":
        return load_model(filepath)

    elif filepath.suffix == ".pickle":
        with open(filepath, "rb") as f:
            return pickle.load(f)

    else:
        raise ValueError("Unknown file format")



from __future__ import annotations

import re

from .physics_object import PhysicsObject


def is_collective(object_: PhysicsObject | str) -> bool:
    """Check if an object is a collective physics object"""
    if isinstance(object_, PhysicsObject):
        return isinstance(object_, Collective)

    return bool(re.match(r"^[a-zA-Z]+$|^[a-zA-Z]+\d*:\d*$", object_))


class Collective(PhysicsObject):
    """A collective physics object"""

    def __init__(
        self,
        branch: str,
        start: int | None = None,
        stop: int | None = None,
    ) -> None:
        self._branch = branch
        self._start = start
        self._stop = stop

    @classmethod
    def from_name(cls, name: str) -> Collective:
        if re.match(r"^[a-zA-Z]+$|^[a-zA-Z]+\d*:\d*$", name.strip()):
            match_ = re.match(r"^([a-zA-Z]+)(\d*):?(\d*)$", name)
            branch, start, stop = match_.groups()
            start = int(start) if start != "" else None
            stop = int(stop) if stop != "" else None

            return cls(branch, start, stop)

        raise ValueError(f"Invalid name '{name}' for a collective physics object")

    @property
    def start(self) -> int | None:
        return self._start

    @property
    def stop(self) -> int | None:
        return self._stop

    @property
    def branch(self) -> str:
        return self._branch

    @property
    def slices(self) -> list[slice]:
        return [slice(self.start, self.stop)]

    @property
    def name(self) -> str:
        if self.start is None and self.stop is None:
            return f"{self.branch}"

        elif self.start is None:
            return f"{self.branch}:{self.stop}"

        elif self.stop is None:
            return f"{self.branch}{self.start}:"

        else:
            return f"{self.branch}{self.start}:{self.stop}"

    @property
    def config(self) -> dict:
        return {"branch": self.branch, "start": self.start, "stop": self.stop}



from __future__ import annotations

from .collective import Collective, is_collective
from .multiple import Multiple, is_multiple
from .nested import Nested, is_nested
from .physics_object import PhysicsObject
from .single import Single, is_single

ALL_OBJECTS = {Single, Collective, Nested, Multiple}
ALL_OBJECTS_DICT = {cls.__name__: cls for cls in ALL_OBJECTS}
ALL_OBJECTS_DICT.update({cls.__name__.lower(): cls for cls in ALL_OBJECTS})


def get(identifier: str) -> PhysicsObject | None:
    """Retrieve a physics object class from its identifier"""
    return ALL_OBJECTS_DICT.get(identifier)


def parse_physics_object(name: str) -> PhysicsObject:
    """Parse a name to create a physics object"""
    if is_single(name):
        return Single.from_name(name)

    elif is_collective(name):
        return Collective.from_name(name)

    elif is_nested(name):
        return Nested.from_name(name)

    elif is_multiple(name):
        return Multiple.from_name(name)

    else:
        raise ValueError(f"Invalid name '{name}' for a physics object")



from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import Any, Literal, NamedTuple

import dill as pickle
import keras
from numpy import ndarray
from numpy.random import RandomState
from sklearn.base import BaseEstimator
from sklearn.ensemble import GradientBoostingClassifier

KERAS_METRICS = {
    "binary_accuracy": keras.metrics.BinaryAccuracy,
    "categorical_accuracy": keras.metrics.CategoricalAccuracy,
    "sparse_categorical_accuracy": keras.metrics.SparseCategoricalAccuracy,
    "top_k_categorical_accuracy": keras.metrics.TopKCategoricalAccuracy,
    "sparse_top_k_categorical_accuracy": keras.metrics.SparseTopKCategoricalAccuracy,
    "binary_crossentropy": keras.metrics.BinaryCrossentropy,
    "category_crossentropy": keras.metrics.CategoricalCrossentropy,
    "sparse_category_crossentropy": keras.metrics.SparseCategoricalCrossentropy,
    "kullback_leibler_divergence": keras.metrics.KLDivergence,
    "poisson": keras.metrics.Poisson,
    "auc": keras.metrics.AUC,
    "precision": keras.metrics.Precision,
    "recall": keras.metrics.Recall,
    "true_positives": keras.metrics.TruePositives,
    "true_negatives": keras.metrics.TrueNegatives,
    "false_positives": keras.metrics.FalsePositives,
    "false_negatives": keras.metrics.FalseNegatives,
    "precision_at_recall": keras.metrics.PrecisionAtRecall,
    "sensitivity_at_specificity": keras.metrics.SensitivityAtSpecificity,
    "specificity_at_sensitivity": keras.metrics.SpecificityAtSensitivity,
}


class History(NamedTuple):
    history: dict[str, Any]


class GradientBoostedDecisionTree(GradientBoostingClassifier):
    def __init__(
        self,
        name: str = "gradient_boosted_decision_tree",
        *,
        loss: Literal["log_loss", "deviance", "exponential"] = "log_loss",
        learning_rate: float = 0.1,
        n_estimators: int = 100,
        subsample: float = 1,
        criterion: Literal["friedman_mse", "squared_error"] = "friedman_mse",
        min_samples_split: float | int = 2,
        min_samples_leaf: float | int = 1,
        min_weight_fraction_leaf: float = 0,
        max_depth: int | None = 3,
        min_impurity_decrease: float = 0,
        init: str | BaseEstimator | None = None,
        random_state: int | RandomState | None = None,
        max_features: float | int | Literal["auto", "sqrt", "log2"] | None = None,
        verbose: int = 0,
        max_leaf_nodes: int | None = None,
        warm_start: bool = False,
        validation_fraction: float = 0.1,
        n_iter_no_change: int | None = None,
        tol: float = 0.0001,
        ccp_alpha: float = 0,
    ) -> None:
        super().__init__(
            loss=loss,
            learning_rate=learning_rate,
            n_estimators=n_estimators,
            subsample=subsample,
            criterion=criterion,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            min_weight_fraction_leaf=min_weight_fraction_leaf,
            max_depth=max_depth,
            min_impurity_decrease=min_impurity_decrease,
            init=init,
            random_state=random_state,
            max_features=max_features,
            verbose=verbose,
            max_leaf_nodes=max_leaf_nodes,
            warm_start=warm_start,
            validation_fraction=validation_fraction,
            n_iter_no_change=n_iter_no_change,
            tol=tol,
            ccp_alpha=ccp_alpha,
        )
        self.name = name
        self.metrics = []

    def compile(self, optimizer=None, loss="log_loss", metrics=None):
        self.optimizer = optimizer
        self.set_params(loss=loss)
        self.metrics = metrics or []

    def fit(
        self,
        x: ndarray,
        y: ndarray,
        validation_split: float = 0.0,
        validation_data: tuple[ndarray, ndarray] | None = None,
        sample_weight: ndarray | None = None,
        verbose: int = 1,
    ):
        is_categorical = y.ndim == 2
        pb = keras.utils.Progbar(self.n_estimators, verbose=verbose)  # type: ignore
        history = History(history=defaultdict(list))

        if validation_data is not None:
            self.x_val, self.y_val = validation_data
        elif validation_split != 0.0:
            self.x_val = x[-int(validation_split * len(x)) :]
            x = x[: -int(validation_split * len(x))]
            self.y_val = y[-int(validation_split * len(y)) :]
            y = y[: -int(validation_split * len(y))]

        self.metric_pairs = []
        for metric in self.metrics:
            if issubclass(type(metric), keras.metrics.Metric):
                self.metric_pairs.append((metric.name, metric))
            elif metric in KERAS_METRICS:
                self.metric_pairs.append((metric, KERAS_METRICS[metric]()))
            elif metric in ["accuracy", "acc"]:
                if is_categorical:
                    self.metric_pairs.append(
                        (metric, keras.metrics.CategoricalAccuracy())
                    )
                else:
                    self.metric_pairs.append(
                        (metric, keras.metrics.SparseCategoricalAccuracy())
                    )
            else:
                raise ValueError(f"Unknown metric: {metric}")

        def _monitor(i, model, local_variables):
            y_true = local_variables["y"]
            raw_pred = local_variables["raw_predictions"]

            if hasattr(model._loss, "predict_proba"):
                # This works in 1.4.2
                y_prob = model._loss.predict_proba(raw_pred)
            else:
                # This works in 1.3.0
                y_prob = model._loss._raw_prediction_to_proba(raw_pred)
            sample_weight = local_variables["sample_weight"]

            loss = model._loss(y_true, raw_pred, sample_weight)

            y_true = (
                y_true if not is_categorical else keras.utils.to_categorical(y_true)
            )

            # Train metrics
            train_values = [("loss", loss)]
            history.history["loss"].append(loss)
            for name, i in self.metric_pairs:
                i.reset_state()
                i.update_state(y_true, y_prob, sample_weight=sample_weight)
                train_values.append((name, i.result().numpy()))
                history.history[name].append(i.result().numpy())

            # Validation metrics
            val_values = []
            if validation_split != 0.0 or validation_data is not None:
                # Reshape y_val and change dtype to match the previous one
                y_true = (
                    self.y_val if self.y_val.ndim == 1 else self.y_val.argmax(axis=1)
                ).astype(y_true.dtype)  # (n_samples,)
                raw_pred = next(model._staged_raw_predict(self.x_val))  # (n_samples, 1)
                if hasattr(model._loss, "predict_proba"):
                    # This works in 1.4.2
                    y_prob = model._loss.predict_proba(raw_pred)
                else:
                    # This works in 1.3.0
                    y_prob = model._loss._raw_prediction_to_proba(raw_pred)
                # y_prob = model._loss._raw_prediction_to_proba(
                #     raw_pred
                # )  # (n_samples, n_classes)

                # ! sample_weight: (n_samples x (1 - validation_fraction),)
                # here the validation_fraction is a parameter of the parent class
                # print(y_true.shape, raw_pred.shape, y_prob.shape, sample_weight.shape)
                val_loss = model._loss(y_true, raw_pred)  # , sample_weight)
                val_values.append(("val_loss", val_loss))
                history.history["val_loss"].append(val_loss)

                for name, i in self.metric_pairs:
                    i.reset_state()
                    i.update_state(self.y_val, y_prob)
                    val_values.append(("val_" + name, i.result().numpy()))
                    history.history["val_" + name].append(i.result().numpy())

            pb.add(1, values=[("loss", loss)] + train_values + val_values)

            return False

        y = y if not is_categorical else y.argmax(axis=1)
        _ = super().fit(x, y, sample_weight, _monitor)
        return history

    def predict(self, x: ndarray, **kwargs) -> ndarray:
        return super().predict_proba(x)

    def summary(self, deep=False, **kwargs):
        output = [f'Model: "{self.name}"']
        for name, value in self.get_params(deep=deep, **kwargs).items():
            output.append(f"- {name}: {value}")

        print("\n".join(output))

    def save(
        self,
        filepath,
        overwrite=True,
        save_format: Literal["pickle"] = "pickle",
    ):
        filepath = Path(filepath)
        if filepath.suffix != "":
            save_format = filepath.suffix[1:]  # type: ignore

        filepath = filepath.with_suffix(f".{save_format}")
        if not overwrite and filepath.exists():
            raise FileExistsError(f"File already exists: {filepath}")

        with filepath.open("wb") as f:
            if save_format == "pickle":
                pickle.dump(self, f)
            else:
                raise ValueError(f"save_format {save_format} not supported")



from .fastjet_ops import get_jet_algorithm
from .keras_ops import ops_histogram_fixed_width, ops_unique
from .uproot_ops import (
    branch_to_momentum4d,
    constituents_to_momentum4d,
    find_eflow_in_refs,
    take_momentum4d,
)



from __future__ import annotations

import json
import zipfile

from .graph_dataset import GraphDataset
from .image_dataset import ImageDataset
from .set_dataset import SetDataset


def load_dataset(filepath, lazy=True):
    zf = zipfile.ZipFile(filepath)
    # Extract and read configs JSON
    with zf.open("configs.json") as json_file:
        configs = json.load(json_file)

    if configs["class_name"] == "SetDataset":
        ds = SetDataset.load(filepath, lazy=lazy)
    else:
        ds = ImageDataset.load(filepath, lazy=lazy)

    return ds



from __future__ import annotations

import awkward as ak
import numba as nb
import vector

vector.register_awkward()


@nb.njit(cache=True)
def find_eflow_in_refs(eflow, refs):  # pragma: no cover
    """Find the constituent reference indices in an eflow branch.

    This operation is used to retrieve the jet constituents. Jet constituents are
    a reference to the "EFlowTrack", "EFlowPhoton", and "EFlowNeutralHadron" branches
    in the Delphes output.

    The reference is a 2d record array with the shape (n, var, var), which means
    n events with variable-length jets and variable-length constituents per jet.
    The corresponding branch, however, is a 1d record array. To related the two,
    we need to find the "fUniqueID" of the jet constituents in the reference array.

    Parameters
    ----------
    eflow: awkward array, shape (n, var)
        EFlowTrack, EFlowPhoton, or EFlowNeutralHadron array.
    refs: awkward array, shape (n, var, var)
        Reference array of all jet constituents.

    Return
    ------
    indices: list
        List of indices of constituents in the eflow array.
    """
    indices = []

    for record_a, record_b in zip(eflow, refs):
        indices_per_a = []

        for i in range(len(record_b)):
            indices_per_b = []

            for j in range(len(record_b[i])):
                for k in range(len(record_a)):
                    if record_b[i][j] == record_a[k]:
                        indices_per_b.append(k)
            indices_per_a.append(indices_per_b)

        indices.append(indices_per_a)

    return indices


@nb.njit(cache=True)
def take_momentum4d(array, indices):  # pragma: no cover
    """Take the 4-momentum of the constituents in the eflow array.

    Parameters
    ----------
    array: awkward array, shape (n, var)
        EFlowTrack, EFlowPhoton, or EFlowNeutralHadron array.
    indices: list
        List of indices of constituents in the eflow array.

    Return
    ------
    momentum4d: tuple
        Tuple of 4-momentum components (pt, eta, phi, mass).
    """
    pt = []
    eta = []
    phi = []
    mass = []

    for record, indices_at_1 in zip(array, indices):
        pt.append(
            [[record[i].pt for i in indices_at_2] for indices_at_2 in indices_at_1]
        )
        phi.append(
            [[record[i].phi for i in indices_at_2] for indices_at_2 in indices_at_1]
        )
        eta.append(
            [[record[i].eta for i in indices_at_2] for indices_at_2 in indices_at_1]
        )
        mass.append(
            [[record[i].mass for i in indices_at_2] for indices_at_2 in indices_at_1]
        )

    return pt, eta, phi, mass


def branch_to_momentum4d(events, branch, with_id=False):
    """Convert a Delphes branch to a 4-momentum array.

    Some branches in the Delphes output do not have full 4-momentum information,
    e.g., "Jet" has PT, Eta, Phi, Mass but no Px, Py, Pz, E. This function converts
    the branch to the registered "Momentum4D" array supported by the vector library.

    Parameters
    ----------
    events:
        Events opened by uproot.
    branch: str
        Branch name to be converted, e.g., "Jet"
    with_id: bool
        Whether to include the "fUniqueID" in the 4-momentum array, which is used
        to find the constituents in the eflow branches.

    Return
    ------
    momenta: Momentum4D
        4-momentum array with the shape (n, var).
    """
    eta = events[f"{branch}.Eta"].array()
    phi = events[f"{branch}.Phi"].array()

    if f"{branch}.PT" in events:
        pt = events[f"{branch}.PT"].array()
    elif f"{branch}.ET" in events:
        pt = events[f"{branch}.ET"].array()
    elif f"{branch}.MET" in events:
        pt = events[f"{branch}.MET"].array()
    else:
        raise ValueError(f"Cannot find the PT branch for {branch}")

    if f"{branch}.Mass" in events:
        mass = events[f"{branch}.Mass"].array()
    else:
        mass = ak.zeros_like(eta)

    momenta = ak.zip(
        {
            "pt": pt,
            "eta": eta,
            "phi": phi,
            "mass": mass,
        },
        with_name="Momentum4D",
    )
    momenta = ak.values_astype(momenta, "float32")

    if with_id:
        momenta["id"] = events[f"{branch}.fUniqueID"].array()

    return momenta


def constituents_to_momentum4d(events, branch):
    """Convert the constituents in a Delphes branch to a 4-momentum array.

    Parameters
    ----------
    events:
        Events opened by uproot.
    branch: str
        Branch name to be converted, e.g., "Jet.Constituents"

    Return
    ------
    constituents: Momentum4D
        4-momentum array with the shape (n, var, var).
    """
    refs = events[branch].array()["refs"]
    tracks = branch_to_momentum4d(events, "EFlowTrack", with_id=True)
    photons = branch_to_momentum4d(events, "EFlowPhoton", with_id=True)
    neutrals = branch_to_momentum4d(events, "EFlowNeutralHadron", with_id=True)

    matches = []
    for i in [tracks, photons, neutrals]:
        indices = ak.from_iter(find_eflow_in_refs(i.id, refs))
        pt, eta, phi, mass = take_momentum4d(i, indices)
        matches.append(
            ak.zip(
                {
                    "pt": pt,
                    "eta": eta,
                    "phi": phi,
                    "mass": mass,
                },
                with_name="Momentum4D",
            )
        )

    constituents = ak.concatenate(matches, -1)
    constituents = ak.values_astype(constituents, "float32")

    return constituents



from __future__ import annotations

import re

from .angular_distance import AngularDistance
from .charge import Charge
from .invariant_mass import InvariantMass
from .kinematics import E, Eta, M, Phi, Pt, Px, Py, Pz
from .n_subjettiness import NSubjettiness, NSubjettinessRatio, TauMN, TauN
from .observable import Observable
from .size import Size
from .tag import BTag, TauTag

ALL_OBJECTS = {
    Px,
    Py,
    Pz,
    E,
    Pt,
    Eta,
    Phi,
    M,
    Charge,
    BTag,
    TauTag,
    Size,
    NSubjettiness,
    NSubjettinessRatio,
    TauMN,
    TauN,
    InvariantMass,
    AngularDistance,
}
ALL_OBJECTS_DICT = {obj.__name__: obj for obj in ALL_OBJECTS}
ALL_OBJECTS_DICT.update({obj.__name__.lower(): obj for obj in ALL_OBJECTS})
ALL_OBJECTS_DICT.update(
    {
        "MomentumX": Px,
        "momentum_x": Px,
        "MomentumY": Py,
        "momentum_y": Py,
        "MomentumZ": Pz,
        "momentum_z": Pz,
        "Energy": E,
        "energy": E,
        "TransverseMomentum": Pt,
        "transverse_momentum": Pt,
        "PT": Pt,
        "met": Pt,
        "MET": Pt,
        "PseudoRapidity": Eta,
        "pseudo_rapidity": Eta,
        "AzimuthalAngle": Phi,
        "azimuthal_angle": Phi,
        "Mass": M,
        "mass": M,
        "b_tag": BTag,
        "tau_tag": TauTag,
        "n_subjettiness": NSubjettiness,
        "n_subjettiness_ratio": NSubjettinessRatio,
        "tau_mn": TauMN,
        "tau_n": TauN,
        "invariant_mass": InvariantMass,
        "InvMass": InvariantMass,
        "inv_mass": InvariantMass,
        "InvM": InvariantMass,
        "inv_m": InvariantMass,
        "angular_distance": AngularDistance,
        "DeltaR": AngularDistance,
        "delta_r": AngularDistance,
    }
)


def get(identifier: str | None) -> Observable | None:
    if identifier is None or identifier == "None":
        return

    else:
        return ALL_OBJECTS_DICT.get(identifier)


def register_observable(obs: Observable, *alias: str) -> None:
    ALL_OBJECTS.add(obs)

    for name in alias:
        ALL_OBJECTS_DICT[name] = obs


def parse_observable(name: str | None, **kwargs) -> Observable | None:
    if name is None or (isinstance(name, str) and name == "None"):
        return

    if (class_name := name.split(".")[-1]) in ALL_OBJECTS_DICT:
        kwargs["class_name"] = class_name
        return ALL_OBJECTS_DICT[class_name].from_name(name, **kwargs)

    elif re.match(r"^tau\d$", class_name.lower()):
        kwargs["class_name"] = class_name
        return TauN.from_name(name, **kwargs)

    elif re.match(r"^tau\d\d$", class_name.lower()):
        kwargs["class_name"] = class_name
        return TauMN.from_name(name, **kwargs)

    else:
        raise ValueError(f"Invalid '{name}' for an observable")



from __future__ import annotations

import fastjet as fj


def get_jet_algorithm(name: str):
    JET_ALGORITHMS = {
        "kt": fj.kt_algorithm,
        "cambridge": fj.cambridge_algorithm,
        "antikt": fj.antikt_algorithm,
        "genkt": fj.genkt_algorithm,
        "cambridge_for_passive": fj.cambridge_for_passive_algorithm,
        "genkt_for_passive": fj.genkt_for_passive_algorithm,
        "ee_kt": fj.ee_kt_algorithm,
        "ee_genkt": fj.ee_genkt_algorithm,
        "plugin": fj.plugin_algorithm,
        "undefined": fj.undefined_jet_algorithm,
    }

    return JET_ALGORITHMS.get(name, fj.undefined_jet_algorithm)



from __future__ import annotations

import awkward as ak

from hml.physics_objects.physics_object import PhysicsObject

from .observable import Observable


class NSubjettiness(Observable):
    def __init__(
        self,
        n: int,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective"]
        super().__init__(physics_object, class_name, supported_objects)
        self.n = n

    def read(self, events):
        all_keys = {i.lower(): i for i in events.keys(full_paths=False)}
        branch = self.physics_object.branch.lower()
        slices = self.physics_object.slices

        if f"{branch}.tau[5]" in all_keys:
            key = all_keys[f"{branch}.tau[5]"]
            array = events[key].array()[:, slices[0]]
            value = array[:, :, self.n - 1]

        else:
            raise ValueError

        for i, slice_ in enumerate(slices):
            if slice_.stop is not None:
                start = slice_.start if slice_.start is not None else 0
                required_length = slice_.stop - start

                if i + 1 == len(slices) and ak.any(
                    ak.num(value, i + 1) < required_length
                ):
                    value = ak.pad_none(value, required_length, axis=i + 1)

                else:
                    n_missing = required_length - ak.num(value, i + 1)
                    if ak.sum(n_missing) > 0:
                        pad = ak.unflatten(
                            ak.Array([[]] * ak.sum(n_missing)), n_missing
                        )
                        value = ak.concatenate([value, pad], axis=i + 1)

        self._value = value

        return self

    @property
    def config(self):
        config = super().config
        config.update({"n": self.n})
        return config


class TauN(NSubjettiness):
    @classmethod
    def from_name(cls, name: str, **kwargs) -> TauN:
        *parts, class_name = name.split(".")
        physics_object = ".".join(parts) if len(parts) > 0 else None

        if class_name.lower().startswith("tau"):
            if "n" not in kwargs:
                n = int(class_name[-1])
            else:
                n = kwargs["n"]

        return cls(n, physics_object, class_name)


class NSubjettinessRatio(Observable):
    def __init__(
        self,
        m: int,
        n: int,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective"]
        super().__init__(physics_object, class_name, supported_objects)
        self.m = m
        self.n = n

        self.tau_m = TauN(m, physics_object)
        self.tau_n = TauN(n, physics_object)

    def read(self, events):
        self.tau_m.read(events)
        self.tau_n.read(events)

        self._value = self.tau_m.value / self.tau_n.value

        return self

    @property
    def config(self):
        config = super().config
        config.update({"m": self.m, "n": self.n})
        return config


class TauMN(NSubjettinessRatio):
    @classmethod
    def from_name(cls, name: str, **kwargs) -> TauMN:
        *parts, class_name = name.split(".")
        physics_object = ".".join(parts) if len(parts) > 0 else None

        if class_name.lower().startswith("tau"):
            if "m" not in kwargs:
                m = int(class_name[-2])
            else:
                m = kwargs["m"]

            if "n" not in kwargs:
                n = int(class_name[-1])
            else:
                n = kwargs["n"]

        return cls(m, n, physics_object, class_name)



from __future__ import annotations

from ..physics_objects import PhysicsObject
from .observable import Observable


class Px(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class Py(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class Pz(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class E(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class Pt(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class Eta(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class Phi(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)


class M(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective", "nested"]
        super().__init__(physics_object, class_name, supported_objects)



from __future__ import annotations

from importlib import import_module

import awkward as ak
import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import vector
from fastjet import ClusterSequence, JetDefinition

from hml.observables import parse_observable
from hml.operations import get_jet_algorithm

vector.register_awkward()


class Image:
    def __init__(self, height, width, channel=None):
        self.height = parse_observable(height) if isinstance(height, str) else height
        self.width = parse_observable(width) if isinstance(width, str) else width

        if channel is not None:
            self.channel = (
                parse_observable(channel) if isinstance(channel, str) else channel
            )
        else:
            self.channel = None

        self.been_pixelated = None
        self.been_read = False
        self.registered_methods = []
        self.recorded_operations = []
        self.status = True

    def read(self, events):
        self.been_pixelated = None
        self.been_read = False
        self.status = True

        self.event = events
        self.height.read(events)
        self.height._value = ak.flatten(self.height._value, axis=-1)
        self.width.read(events)
        self.width._value = ak.flatten(self.width._value, axis=-1)

        if self.channel is not None:
            self.channel.read(events)
            self.channel._value = ak.flatten(self.channel._value, axis=-1)
        self.been_read = True

        for method, kwargs in self.registered_methods:
            getattr(self, method)(**kwargs)

        return self

    def with_subjets(self, constituents, algorithm, r, min_pt):
        if self.been_read:
            px = parse_observable(f"{constituents}.Px").read(self.event).value
            py = parse_observable(f"{constituents}.Py").read(self.event).value
            pz = parse_observable(f"{constituents}.Pz").read(self.event).value
            e = parse_observable(f"{constituents}.E").read(self.event).value

            px = ak.flatten(px, -1)
            py = ak.flatten(py, -1)
            pz = ak.flatten(pz, -1)
            e = ak.flatten(e, -1)

            particles = ak.zip(
                {"px": px, "py": py, "pz": pz, "e": e}, with_name="Momentum4D"
            )
            subjet_def = JetDefinition(get_jet_algorithm(algorithm), r)
            self.cluster = ClusterSequence(particles, subjet_def)
            self.subjets = self.cluster.inclusive_jets(min_pt)

            self.recorded_operations.append(
                (
                    "with_subjets",
                    {
                        "constituents": constituents,
                        "algorithm": algorithm,
                        "r": r,
                        "min_pt": min_pt,
                    },
                )
            )
        else:
            self.registered_methods.append(
                (
                    "with_subjets",
                    {
                        "constituents": constituents,
                        "algorithm": algorithm,
                        "r": r,
                        "min_pt": min_pt,
                    },
                )
            )

        return self

    def translate(self, origin="SubJet0"):
        if self.been_read:
            origin_height = parse_observable(
                f"{origin}.{self.height.__class__.__name__}"
            )
            origin_width = parse_observable(f"{origin}.{self.width.__class__.__name__}")

            obj = origin_height.physics_object.branch
            index = origin_height.physics_object.index

            if obj != "SubJet":
                raise ValueError(f"{obj} is not supported yet!")

            if len(self.subjets) < index + 1:
                self.status = False
                return self

            origin_height._value = ak.to_regular(
                ak.pad_none(
                    getattr(self.subjets, origin_height.__class__.__name__.lower())[
                        :, : index + 1
                    ],
                    1,
                )
            )

            origin_width._value = ak.to_regular(
                ak.pad_none(
                    getattr(self.subjets, origin_width.__class__.__name__.lower())[
                        :, : index + 1
                    ],
                    1,
                )
            )

            self.origin_height = origin_height
            self.origin_width = origin_width

            translated_height = self.height.value - origin_height.value
            if self.height.__class__.__name__ == "Phi":
                translated_height = np.mod(translated_height + np.pi, 2 * np.pi) - np.pi

            translated_width = self.width.value - origin_width.value
            if self.width.__class__.__name__ == "Phi":
                translated_width = np.mod(translated_width + np.pi, 2 * np.pi) - np.pi

            self.height._value = translated_height
            self.width._value = translated_width

            self.recorded_operations.append(("translate", {"origin": origin}))
        else:
            self.registered_methods.append(("translate", {"origin": origin}))

        return self

    def rotate(self, axis="SubJet1", orientation=-90):
        if self.been_read:
            axis_height = parse_observable(f"{axis}.{self.height.__class__.__name__}")
            axis_width = parse_observable(f"{axis}.{self.width.__class__.__name__}")

            obj = axis_height.physics_object.branch
            index = axis_height.physics_object.index

            if obj != "SubJet":
                raise ValueError(f"{obj} is not supported yet!")

            if len(self.subjets) < index + 1:
                self.status = False
                return self

            axis_height._value = ak.to_regular(
                ak.pad_none(
                    getattr(self.subjets, axis_height.__class__.__name__.lower())[
                        :, index : index + 1
                    ],
                    1,
                )
            )
            axis_width._value = ak.to_regular(
                ak.pad_none(
                    getattr(self.subjets, axis_width.__class__.__name__.lower())[
                        :, index : index + 1
                    ],
                    1,
                )
            )

            delta_h = axis_height.value - self.origin_height.value
            delta_w = axis_width.value - self.origin_width.value
            angle = orientation - np.arctan2(delta_h, delta_w) * 180 / np.pi
            angle = np.deg2rad(angle)

            # rotation_matrix = np.array(
            #     [
            #         [np.cos(angle), -np.sin(angle)],
            #         [np.sin(angle), np.cos(angle)],
            #     ]
            # )  # (2, 2)
            # points = np.vstack(
            #     [self.width.to_numpy(), self.height.to_numpy()]
            # )  # (2, n)
            # rotated_points = np.dot(rotation_matrix, points)  # (2, n)
            angle = ak.values_astype(ak.fill_none(angle, np.nan), "float32")
            x = self.width.value
            y = self.height.value
            x_prime = np.cos(angle) * x - np.sin(angle) * y
            y_prime = np.sin(angle) * x + np.cos(angle) * y

            self.width._value = x_prime
            self.height._value = y_prime

            self.recorded_operations.append(
                ("rotate", {"axis": axis, "orientation": orientation})
            )

        else:
            self.registered_methods.append(
                ("rotate", {"axis": axis, "orientation": orientation})
            )

        return self

    def pixelate(self, size, range):
        self.been_pixelated = True
        self.h_bins = np.linspace(*range[0], size[0] + 1)
        self.w_bins = np.linspace(*range[1], size[1] + 1)

        if self.been_read:
            if self.status is False:
                return self

            self.height._value = self.continuous_to_center(
                self.height.value, self.h_bins
            )
            # pixelated_values = self.continuous_to_center(
            #     self.height.to_numpy(), self.h_bins
            # )
            # pixelated_values = pixelated_values.reshape(
            #     self.height.to_numpy(squeeze=False).shape
            # )
            # self.height._value = pixelated_values.tolist()

            self.width._value = self.continuous_to_center(self.width.value, self.w_bins)

            # pixelated_values = self.continuous_to_center(
            #     self.width.to_numpy(), self.w_bins
            # )
            # pixelated_values = pixelated_values.reshape(
            #     self.width.to_numpy(squeeze=False).shape
            # )
            # self.width._value = pixelated_values.tolist()

            self.recorded_operations.append(
                ("pixelate", {"size": size, "range": range})
            )

        else:
            self.registered_methods.append(("pixelate", {"size": size, "range": range}))

        return self

    def continuous_to_center(self, values, bins):
        def _transform_func(layout, **kwargs):
            if layout.is_numpy:
                out_values = np.empty_like(layout.data)
                bin_centers = (bins[:-1] + bins[1:]) / 2
                bin_indices = np.digitize(layout.data, bins)

                for i, index in enumerate(bin_indices):
                    if index == 0 or index == len(bins):
                        out_values[i] = np.nan
                    else:
                        out_values[i] = bin_centers[index - 1]

                return ak.contents.NumpyArray(out_values)

        return ak.transform(_transform_func, values)

    @property
    def values(self):
        widths = ak.fill_none(self.width.value, np.nan)
        heights = ak.fill_none(self.height.value, np.nan)
        total = len(widths)

        if self.been_pixelated is not None and self.been_read:
            if self.channel is not None:
                weights = ak.fill_none(self.channel.value, np.nan)
                hist = calculate_histograms(
                    widths, heights, self.w_bins, self.h_bins, total, weights
                )
            else:
                hist = calculate_histograms(
                    widths, heights, self.w_bins, self.h_bins, total
                )
            return hist

        return self.height.value, self.width.value

    def show(
        self,
        as_point=False,
        limits=None,
        show_pixels=False,
        grid=True,
        norm=None,
    ):
        plt.figure()

        if not self.been_pixelated:
            plt.scatter(
                x=ak.flatten(self.width.value, None),
                y=ak.flatten(self.height.value, None),
                c="k",
                s=10,
                marker="o",
            )

            # Restrict the range of axes.
            if limits is not None:
                plt.xlim(limits[0])
                plt.ylim(limits[1])

        elif as_point:
            plt.scatter(
                x=ak.flatten(self.width.value, None),
                y=ak.flatten(self.height.value, None),
                c="k",
                s=10,
                marker="s",
            )
            plt.xlim(self.w_bins[0], self.w_bins[-1])
            plt.ylim(self.h_bins[0], self.h_bins[-1])

            plt.gca().set_aspect("equal")

        else:
            # Use pcolormesh instead of imshow so that the actual values of bins
            # are shown.
            plt.pcolormesh(self.w_bins, self.h_bins, self.values.sum(0).T, norm=norm)

            # Turn on colorbar to show the range of values.
            plt.colorbar()

            if show_pixels:
                # Turn off ticks and labels.
                plt.tick_params(
                    bottom=False, left=False, labelbottom=False, labelleft=False
                )
                plt.xticks(self.w_bins)
                plt.yticks(self.h_bins)

            # Set aspect ratio to be equal so that every pixel is square.
            # Ignore the actual values of bins since here is just a schematic.
            plt.gca().set_aspect("equal")

        # Turn on grid to show pixel boundaries.
        if grid:
            plt.grid(alpha=0.5)

        plt.show()

    @property
    def config(self):
        return {
            "height_config": {
                "class_name": self.height.__class__.__name__,
                "config": self.height.config,
            },
            "width_config": {
                "class_name": self.width.__class__.__name__,
                "config": self.width.config,
            },
            "channel_config": (
                {
                    "class_name": self.channel.__class__.__name__,
                    "config": self.channel.config if self.channel is not None else None,
                }
                if self.channel is not None
                else None
            ),
            "registered_methods": (
                self.registered_methods
                if not self.been_read
                else self.recorded_operations
            ),
            "been_pixelated": self.been_pixelated,
            "w_bins": (self.w_bins.tolist() if self.been_pixelated else None),
            "h_bins": (self.h_bins.tolist() if self.been_pixelated else None),
        }

    @classmethod
    def from_config(cls, config):
        module = import_module("hml.observables")

        height_class_name = config["height_config"]["class_name"]
        height_class = getattr(module, height_class_name)
        height = height_class.from_config(config["height_config"]["config"])

        width_class_name = config["width_config"]["class_name"]
        width_class = getattr(module, width_class_name)
        width = width_class.from_config(config["width_config"]["config"])

        if config["channel_config"] is not None:
            channel_class_name = config["channel_config"]["class_name"]
            channel_class = getattr(module, channel_class_name)
            channel = channel_class.from_config(config["channel_config"]["config"])
        else:
            channel = None

        instance = cls(height, width, channel)

        instance.registered_methods = config["registered_methods"]
        instance.been_pixelated = config["been_pixelated"]
        instance.w_bins = (
            np.array(config["w_bins"]) if config["w_bins"] is not None else None
        )
        instance.h_bins = (
            np.array(config["h_bins"]) if config["h_bins"] is not None else None
        )

        return instance


@nb.njit
def histogram2d_numba_weighted(x, y, bins, xrange, yrange, weights):
    hist = np.zeros((bins[0], bins[1]), dtype=np.float64)

    x_bin_width = (xrange[1] - xrange[0]) / bins[0]
    y_bin_width = (yrange[1] - yrange[0]) / bins[1]

    for i in range(len(x)):
        if xrange[0] <= x[i] < xrange[1] and yrange[0] <= y[i] < yrange[1]:
            x_bin = int((x[i] - xrange[0]) / x_bin_width)
            y_bin = int((y[i] - yrange[0]) / y_bin_width)

            # Ensure the indices are within the range of the histogram
            x_bin = min(x_bin, bins[0] - 1)
            y_bin = min(y_bin, bins[1] - 1)

            hist[x_bin, y_bin] += weights[i]

    return hist


@nb.njit
def histogram2d_numba(x, y, bins, xrange, yrange):
    """
    A simplified version of numpy.histogram2d compatible with Numba.

    Parameters:
    - x, y: Arrays of the same length containing the x and y coordinates of the points.
    - bins: A tuple (x_bins, y_bins) specifying the number of bins in each dimension.
    - xrange: A tuple (x_min, x_max) specifying the minimum and maximum range on the x-axis.
    - yrange: A tuple (y_min, y_max) specifying the minimum and maximum range on the y-axis.

    Returns:
    - A 2D histogram array.
    """
    hist = np.zeros((bins[0], bins[1]), dtype=np.float64)

    x_bin_width = (xrange[1] - xrange[0]) / bins[0]
    y_bin_width = (yrange[1] - yrange[0]) / bins[1]

    for i in range(len(x)):
        if xrange[0] <= x[i] < xrange[1] and yrange[0] <= y[i] < yrange[1]:
            x_bin = int((x[i] - xrange[0]) / x_bin_width)
            y_bin = int((y[i] - yrange[0]) / y_bin_width)

            # Ensure the indices are within the range of the histogram
            x_bin = min(x_bin, bins[0] - 1)
            y_bin = min(y_bin, bins[1] - 1)

            hist[x_bin, y_bin] += 1

    return hist


@nb.njit
def calculate_histograms(widths, heights, w_bins, h_bins, total, weights=None):
    # Assuming widths and heights are flat NumPy arrays of the same length
    # and that missing data has been handled prior to this call.

    hists = np.zeros(
        (total, len(w_bins) - 1, len(h_bins) - 1)
    )  # Pre-allocate based on expected sizes
    if weights is not None:
        for i in range(total):
            hist = histogram2d_numba_weighted(
                widths[i],  # Assuming these are now slices or individual arrays
                heights[i],
                bins=(len(w_bins) - 1, len(h_bins) - 1),
                xrange=(w_bins[0], w_bins[-1]),
                yrange=(h_bins[0], h_bins[-1]),
                weights=weights[i],
            )
            hists[i] = hist
    else:
        for i in range(total):
            hist = histogram2d_numba(
                widths[i],  # Assuming these are now slices or individual arrays
                heights[i],
                bins=(len(w_bins) - 1, len(h_bins) - 1),
                xrange=(w_bins[0], w_bins[-1]),
                yrange=(h_bins[0], h_bins[-1]),
            )
            hists[i] = hist

    return hists



from __future__ import annotations

from keras import ops


def ops_histogram_fixed_width(values, value_range, nbins, dtype="int32"):
    value_min, value_max = value_range
    bin_edges = ops.linspace(value_min, value_max, nbins + 1)
    lower = bin_edges[:-1]  # type: ignore
    upper = bin_edges[1:]  # type: ignore

    return ops.fori_loop(
        0,
        nbins,
        lambda i, s: ops.scatter_update(
            s,
            [[i]],
            [
                ops.count_nonzero(
                    ops.where(
                        ops.logical_and(lower[i] <= values, values <= upper[i]),
                        1.0,
                        0.0,
                    )
                )
            ],
        ),
        ops.zeros((nbins,), dtype=dtype),
    )


def ops_unique(tensor):
    sorted_tensor, sorted_indices = ops.sort(tensor), ops.argsort(tensor)
    selection = ops.not_equal(sorted_tensor[1:], sorted_tensor[:-1])  # type: ignore
    selection = ops.add(ops.squeeze(ops.where(selection), 0), 1)
    unique_indices = ops.append([0], selection)

    try:
        unique_elements = ops.take(sorted_tensor, unique_indices)
        unique_indices = ops.take(sorted_indices, unique_indices)
        return unique_elements

    except Exception:
        return ops.array([0])



from __future__ import annotations

import re

import awkward as ak

from hml.observables import parse_observable


class Cut:
    def __init__(self, expression) -> None:
        self._expression = expression
        self._parse_expression(expression)

    def _parse_expression(self, expression):
        expr = expression.strip()

        self._is_veto = False
        if expr.startswith("veto"):
            self._is_veto = True
            expr = expr.replace("veto", "").strip()

        self._is_any = False
        if expr.startswith("any"):
            self._is_any = True
            expr = expr.replace("any", "").strip()

        expr = expr.replace("and", "&").replace("or", "|")

        # Split the expression by logical operators "and" and "or"
        cuts = expr.split("&")
        cuts = [cut.strip().split("|") for cut in cuts]
        cuts = [c.strip() for cut in cuts for c in cut]
        cuts = [cut.replace("(", "").replace(")", "") for cut in cuts]

        obs_pattern = r"\b(?!\d+\b)(?!\d*\.\d+\b)\S+\b"
        cuts_dict = {}
        for cut in cuts:
            obs = re.findall(obs_pattern, cut)[0]
            if "" not in cut.split(obs):
                new_cut = [cut.split(obs)[0] + obs, obs + cut.split(obs)[1]]
                cuts_dict[new_cut[0]] = obs
                cuts_dict[new_cut[1]] = obs
                expr = expr.replace(cut, f"({new_cut[0]} & {new_cut[1]})")
            else:
                cuts_dict[cut] = obs

        self._cuts_dict = cuts_dict
        self._observables_dict = {
            obs: parse_observable(obs) for obs in cuts_dict.values()
        }
        self._expr = expr

    def read(self, events):
        for obs in self._observables_dict.values():
            obs.read(events)
        observables_dict = self._observables_dict

        cuts_results = {}
        for cut, obs in self._cuts_dict.items():
            result = eval(cut.replace(obs, f"observables_dict['{obs}'].value"))
            cuts_results[cut] = result

        # Validate the type
        shapes = set([obs.shape for obs in observables_dict.values()])
        if len(shapes) != 1:
            raise ValueError

        expression = self._expr
        for cut in cuts_results:
            cut_pattern = r"\b" + cut.replace(".", "\\.") + r"\b"
            expression = re.sub(cut_pattern, f"cuts_results['{cut}']", expression)

        self._value = ak.fill_none(eval(expression), False)

        if self._value.ndim > 1:
            if self._is_any:
                self._value = ak.any(self._value, axis=1)
            else:
                self._value = ak.all(self._value, axis=1)

        if self._is_veto:
            self._value = ~self._value

        return self

    @property
    def value(self):
        return self._value

    @property
    def expression(self):
        return self._expression



from __future__ import annotations

import re
import shutil
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Union

import pexpect
from bs4 import BeautifulSoup
from rich.console import Console
from rich.table import Table

PathLike = Union[str, Path]


class Madgraph5:
    def __init__(
        self,
        executable: PathLike,
        verbose: int = 1,
    ):
        self.executable = executable
        self.child = pexpect.spawn(f"{self.executable.as_posix()}")
        self.process_log = ""
        self.DEFAULT_LOG_DIR = Path("Logs")
        self.DEFAULT_DIAGRAM_DIR = Path("Diagrams")
        self.verbose = verbose

        self.process_log += self.run_command("")

    @property
    def executable(self) -> Path:
        return self._executable

    @executable.setter
    def executable(self, value: PathLike):
        if (_executable := shutil.which(value)) is None:
            raise FileNotFoundError(f"Could not find Madgraph5 executable {value}")

        self._executable = Path(_executable).resolve()

    @property
    def home(self) -> Path:
        return self.executable.parent.parent

    @property
    def version(self) -> str:
        _version = "unknown"
        with (self.home / "VERSION").open() as f:
            for line in f.readlines():
                if line.startswith("version") and _version == "unknown":
                    _version = line.split("=")[1].strip()

        return _version

    def __repr__(self):
        return f"Madgraph5 v{self.version}"

    def run_command(
        self,
        command: str,
        start_marker: str = r"\r\n",
        end_marker: str = r"MG5_aMC>$",
        timeout: int | None = None,
    ) -> str:
        output = ""
        self.child.sendline(command)
        while True:
            if self.child.expect([start_marker, end_marker], timeout) == 1:  # type: ignore
                break

            middle_output = self.child.before.decode()
            output += middle_output + "\r\n"
            if self.verbose > 0:
                print(middle_output)

        self.clean_pypy()
        return output

    def clean_pypy(self):
        py_py = Path("py.py")
        if py_py.exists():
            py_py.unlink()

    def import_model(self, model: PathLike):
        self.process_log += self.run_command(f"import model {model}")

    def define(self, expression: str):
        self.process_log += self.run_command(f"define {expression}")

    def generate(self, *processes: str):
        self._processes = [i for i in processes]
        self.process_log += self.run_command(f"generate {processes[0]}")
        for process in processes[1:]:
            self.process_log += self.run_command(f"add process {process}")

    @property
    def processes(self) -> list[str]:
        # 1st case: generate() has been called
        try:
            return self._processes
        except AttributeError:
            pass

        # 2nd case: from_output() has been called
        try:
            crossx = self.output_dir / "crossx.html"
            with crossx.open() as f:
                soup = BeautifulSoup(f, "html.parser")
                title_col = soup.find_all("h2")[0].text.strip()
                processes = re.findall(r"^Results in the .+ for (.+)", title_col)[0]
                return processes.split(",")
        except AttributeError:
            raise AttributeError("No processes defined yet")

    def display_diagrams(
        self, diagram_dir: PathLike = "Diagrams", overwrite: bool = True
    ):
        self.diagram_dir = Path(diagram_dir)
        if self.diagram_dir.exists():
            if not overwrite:
                raise FileExistsError(f"{self.diagram_dir} already exists")
            else:
                shutil.rmtree(self.diagram_dir)

        self.diagram_dir.mkdir(parents=True)
        self.process_log += self.run_command(f"display diagrams {self.diagram_dir}")

        for eps in self.diagram_dir.glob("*.eps"):
            subprocess.run(f"ps2pdf {eps} {eps.with_suffix('.pdf')}", shell=True)

    def output(self, output_dir: PathLike | None = None, overwrite: bool = True):
        if output_dir is None:
            log = self.run_command("output")
            self.process_log += log

            match = re.findall(r"Output to directory (.+) done.", log)
            self.output_dir = Path(match[0]).resolve()
        else:
            self.output_dir = Path(output_dir).resolve()
            if self.output_dir.exists():
                if overwrite:
                    shutil.rmtree(self.output_dir)
                else:
                    raise FileExistsError(f"{self.output_dir} already exists")
            self.process_log += self.run_command(f"output {self.output_dir}")

            # try:
            #     if self.diagram_dir.exists():
            #         self.diagram_dir = self.diagram_dir.rename(self.DEFAULT_DIAGRAM_DIR)
            #         shutil.move(self.diagram_dir, self.output_dir / self.diagram_dir)
            # except AttributeError:
            self.display_diagrams(self.output_dir / self.DEFAULT_DIAGRAM_DIR)

        self.log_dir = self.output_dir / self.DEFAULT_LOG_DIR
        self.log_dir.mkdir()
        process_log_file = self.log_dir / "process.log"
        process_log_file = process_log_file.resolve()

        with process_log_file.open("w") as f:
            f.write(self.process_log)
        if self.verbose > 0:
            if process_log_file.is_relative_to(Path.cwd()):
                print(
                    "Process log saved to",
                    process_log_file.relative_to(Path.cwd()),
                )
            else:
                print("Process log saved to", process_log_file)

    def launch(
        self,
        shower="off",
        detector="off",
        madspin="off",
        settings={},
        decays=[],
        cards=[],
        multi_run=1,
        seed=None,
        dry=False,
    ):
        run_log = ""

        # In the middle: MadEvent CLI ends with '>'
        commands = f"launch -i {self.output_dir}\n"
        if multi_run == 1:
            commands += "generate_events\n"
        else:
            commands += f"multi_run {multi_run}\n"

        commands += f"shower={shower}\n"
        commands += f"detector={detector}\n"
        commands += f"madspin={madspin}\n"
        commands += "done\n"

        if seed is not None:
            settings["iseed"] = seed
        if settings != {}:
            commands += "\n".join([f"set {k} {v}" for k, v in settings.items()])
            commands += "\n"

        if decays != []:
            commands += "\n".join([f"decay {i}" for i in decays]) + "\n"

        default_pythia8_card = self.output_dir / "Cards/pythia8_card_default.dat"
        default_delphes_card = self.output_dir / "Cards/delphes_card_default.dat"
        resolved_cards = []
        if seed is not None:
            if shower == "on" or shower == "pythia8":
                pythia8_card = None
                for card in cards:
                    if "pythia8" in card:
                        pythia8_card = card

                if pythia8_card is None:
                    pythia8_card = default_pythia8_card

                    with NamedTemporaryFile(
                        delete=False, prefix="pythia8_card_"
                    ) as temp:
                        temp.write(pythia8_card.read_text().encode())
                        temp.write(b"\n! Modified by hep-ml-lab")
                        temp.write(
                            f"\nRandom:setSeed = on\nRandom:seed = {seed}\n".encode()
                        )

                else:
                    with open(pythia8_card, "r") as f:
                        lines = f.readlines()
                        set_seed = False
                        found_seed = False
                        for i, line in enumerate(lines):
                            if line.startswith("Random:setSeed = on"):
                                set_seed = True
                            if line.startswith("Random:seed"):
                                found_seed = True
                                lines[i] = f"Random:seed = {seed}\n"
                                break

                        if not set_seed:
                            lines.append("Random:setSeed = on\n")
                            lines.append(f"Random:seed = {seed}\n")
                        elif not found_seed:
                            lines.append(f"Random:seed = {seed}\n")

                    with NamedTemporaryFile(
                        delete=False, prefix="pythia8_card_"
                    ) as temp:
                        temp.write("".join(lines).encode())

                    resolved_cards.append(temp.name)

            if detector == "on" or detector == "delphes":
                delphes_card = None
                for card in cards:
                    if "delphes" in card:
                        delphes_card = card

                if delphes_card is None:
                    delphes_card = default_delphes_card

                    with NamedTemporaryFile(
                        delete=False, prefix="delphes_card_"
                    ) as temp:
                        temp.write(f"set RandomSeed {seed}\n".encode())
                        temp.write(delphes_card.read_text().encode())

                else:
                    with open(delphes_card, "r") as f:
                        lines = f.readlines()
                        found_seed = False
                        for i, line in enumerate(lines):
                            if line.startswith("set RandomSeed"):
                                found_seed = True
                                lines[i] = f"set RandomSeed {seed}\n"
                                break

                        if not found_seed:
                            lines.insert(0, f"set RandomSeed {seed}\n")

                    with NamedTemporaryFile(
                        delete=False, prefix="delphes_card_"
                    ) as temp:
                        temp.write("".join(lines).encode())

                resolved_cards.append(temp.name)

        if resolved_cards != []:
            commands += "\n".join(resolved_cards) + "\n"
        else:
            commands += "\n".join(cards) + "\n"
        commands += "done\n"

        if dry:
            return commands

        run_log += self.run_command(commands, end_marker=r">$")

        # In the end: Back to Madgraph CLI
        run_log += self.run_command("exit")

        run_name = re.findall(r"survey  (.+) \r\n", run_log)[0]
        run_log_file = self.log_dir / f"{run_name}.log"
        run_log_file = run_log_file.resolve()
        with run_log_file.open("w") as f:
            f.write(run_log)

        if self.verbose > 0:
            if run_log_file.is_relative_to(Path.cwd()):
                print("Run log saved to", run_log_file.relative_to(Path.cwd()))
            else:
                print("Run log saved to", run_log_file)

    @property
    def runs(self) -> list[Madgraph5Run]:
        run_paths = []
        for i in self.output_dir.glob("Events/run_*"):
            if i.is_dir() and i.name.count("_") == 1:
                run_paths.append(i)

        # Sort the runs by their number
        run_paths = sorted(run_paths, key=lambda x: int(x.name.split("_")[-1]))
        runs = [Madgraph5Run(self.output_dir, i.name) for i in run_paths]

        return runs

    def summary(self):
        console = Console()

        if self.output_dir.is_relative_to(Path.cwd()):
            output_dir = self.output_dir.relative_to(Path.cwd())
        else:
            output_dir = self.output_dir
        table = Table(title="\n".join(self.processes), caption=f"Output: {output_dir}")

        table.add_column("#", justify="right")
        table.add_column("Name")
        table.add_column("Collider")
        table.add_column("Tag")
        table.add_column("Cross section (pb)", justify="center")
        table.add_column("N events", justify="right")
        table.add_column("Seed", justify="right")

        for i, run in enumerate(self.runs):
            table.add_row(
                f"{i}",
                f"{run.name}[{len(run.sub_runs)}]",
                f"{run.collider}",
                f"{run.tag}",
                f"{run.cross:.3e} +- {run.error:.3e}",
                f"{run.n_events:,}",
                f"{run.seed}",
            )

        console.print(table)

    @classmethod
    def from_output(cls, output_dir: PathLike, executable: PathLike) -> Madgraph5:
        output_dir = Path(output_dir)
        if not output_dir.exists():
            raise FileNotFoundError(f"Output directory {output_dir} does not exist")

        output_dir = output_dir.resolve()
        mg5 = cls(executable, verbose=0)
        mg5.verbose = 1
        mg5.output_dir = output_dir

        return mg5


class Madgraph5Run:
    def __init__(self, output_dir: PathLike, name: str):
        self.output_dir = Path(output_dir)
        self._events_dir = self.output_dir / "Events"
        self._run_dir = self._events_dir / name

        self._info = self._get_info(output_dir, name)
        self._subs = []
        for i in self._events_dir.glob(f"{name}_*"):
            if i.is_dir():
                self._subs.append(i)

    @property
    def directory(self) -> Path:
        return self._run_dir

    @property
    def name(self) -> str:
        return self._info["name"]

    @property
    def collider(self) -> str:
        return self._info["collider"]

    @property
    def tag(self) -> str:
        return self._info["tag"]

    @property
    def seed(self) -> int:
        return self._info["seed"]

    @property
    def cross(self) -> float:
        return self._info["cross"]

    @property
    def error(self) -> float:
        return self._info["error"]

    @property
    def n_events(self) -> int:
        return self._info["n_events"]

    @property
    def sub_runs(self) -> list[Madgraph5Run]:
        return [Madgraph5Run(self.output_dir, i.name) for i in self._subs]

    def events(self, file_format="root"):  # type: ignore
        if file_format == "root":
            root_files = []
            # events = ROOT.TChain("Delphes")  # type: ignore
            if self.sub_runs != []:
                for run in self.sub_runs:
                    for root_file in run.directory.glob("*.root"):
                        root_files.append(f"{root_file.as_posix()}:Delphes")
            else:
                for root_file in self.directory.glob("*.root"):
                    root_files.append(f"{root_file.as_posix()}:Delphes")

            # keys = uproot.open(root_files[0]).keys()
            # keys = [key for key in keys if "fBits" not in key]

            # events = uproot.concatenate(root_files, filter_name=keys)
            events = root_files
        else:
            raise ValueError(f"File format {file_format} not supported yet.")

        return events

    def __repr__(self) -> str:
        head = self.name
        if self.sub_runs != []:
            head += f" ({len(self.sub_runs)} sub runs)"
        return (
            f"Madgraph5Run {head}:\n"
            f"- collider: {self.collider}\n"
            f"- tag: {self.tag}\n"
            f"- seed: {self.seed}\n"
            f"- cross: {self.cross}\n"
            f"- error: {self.error}\n"
            f"- n_events: {self.n_events}"
        )

    def _get_info(self, output_dir: PathLike, name: str):
        output_dir = Path(output_dir)
        events_dir = output_dir / "Events"
        run_dir = events_dir / name

        if list(run_dir.glob("*_banner.txt")) != []:
            banner_file = list(run_dir.glob("*_banner.txt"))[0]
        elif list(events_dir.glob(f"{name}_banner.txt")) != []:
            banner_file = list(events_dir.glob(f"{name}_banner.txt"))[0]
        else:
            raise FileNotFoundError("Banner file not found")

        crossx_file = output_dir / "crossx.html"

        with crossx_file.open() as f:
            soup = BeautifulSoup(f, "html.parser")

        table = soup.find("table")
        run = {}
        for row in table.find_all("tr")[1:]:  # type: ignore
            columns = row.find_all("td")
            if columns[0].text != name:
                continue

            # Name
            run["name"] = name

            # Collider
            collider_col = columns[1].text.split()
            lpp1, lpp2 = collider_col[:2]
            ebeam1, ebeam2 = collider_col[2], collider_col[4]
            run["collider"] = f"{lpp1}{lpp2}:{ebeam1}x{ebeam2}"

            # Banner
            banner_col = columns[2].text.split()
            run["tag"] = banner_col[0]

            with banner_file.open() as f:
                for line in f.readlines():
                    if "iseed" in line:
                        run["seed"] = int(line.split("=")[0].strip())
                        break

            # Cross section and error
            cross_col = columns[3].text.split()
            cross = float(cross_col[0])
            error = float(cross_col[2])
            run["cross"] = cross
            run["error"] = error

            # N Events
            events_col = columns[4].text.split()
            run["n_events"] = int(events_col[0])

            # ROOT file path
            run["events"] = {}
            if list(run_dir.glob("*lhe*")) != []:
                run["events"]["lhe"] = list(run_dir.glob("*lhe*"))[0].as_posix()
            if list(run_dir.glob("*hepmc*")) != []:
                run["events"]["hepmc"] = list(run_dir.glob("*hepmc*"))[0].as_posix()
            if list(run_dir.glob("*.root")) != []:
                run["events"]["root"] = list(run_dir.glob("*.root"))[0].as_posix()

        return run



from __future__ import annotations

import keras
from keras import Model
from keras.layers import (
    Conv2D,
    Dense,
    Dropout,
    GlobalAveragePooling2D,
    Input,
    MaxPooling2D,
)


@keras.saving.register_keras_serializable()
class SimpleCNN(Model):
    def __init__(self, input_shape, name="simple_cnn", **kwargs):
        super().__init__(name=name, **kwargs)
        self.input_shape = input_shape
        self.conv1 = Conv2D(8, 3, padding="same", activation="relu")
        self.conv2 = Conv2D(16, 3, padding="same", activation="relu")
        self.conv3 = Conv2D(32, 3, padding="same", activation="relu")
        self.max_pool = MaxPooling2D()
        self.global_avg_pool = GlobalAveragePooling2D()
        self.dropout = Dropout(0.5)
        self.dense1 = Dense(2, activation="relu")
        self.dense2 = Dense(2, activation="softmax")

        self.call(Input(shape=input_shape))

    def call(self, x):
        x = self.conv1(x)
        x = self.max_pool(x)
        x = self.conv2(x)
        x = self.max_pool(x)
        x = self.conv3(x)
        x = self.max_pool(x)
        x = self.global_avg_pool(x)
        x = self.dense1(x)
        return self.dense2(x)

    def get_config(self):
        base_config = super().get_config()
        config = {"input_shape": self.input_shape}
        return {**base_config, **config}



from __future__ import annotations

from ..physics_objects import PhysicsObject
from .observable import Observable


class Charge(Observable):
    def __init__(
        self,
        physics_object: str | PhysicsObject,
        class_name: str | None = None,
    ) -> None:
        supported_objects = ["single", "collective"]
        super().__init__(physics_object, class_name, supported_objects)



from __future__ import annotations

from abc import ABC, abstractmethod


class PhysicsObject(ABC):
    def __eq__(self, other: PhysicsObject | str) -> bool:
        if isinstance(other, PhysicsObject):
            return self.name.lower() == other.name.lower()

        return self.name.lower() == other.lower()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        configs = []
        for key, value in self.config.items():
            if isinstance(value, str):
                configs.append(f"{key}='{value}'")
            else:
                configs.append(f"{key}={value}")
        configs = ", ".join(configs)

        return f"{self.__class__.__name__}({configs})"

    @property
    @abstractmethod
    def branch(self) -> str | list[str]: ...

    @property
    @abstractmethod
    def slices(self) -> list[slice] | list[list[slice]]: ...

    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def config(self) -> dict: ...

    @classmethod
    @abstractmethod
    def from_name(cls, name: str) -> PhysicsObject: ...

    @classmethod
    def from_config(cls, config: dict) -> PhysicsObject:
        return cls(**config)
